#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .simulation import *


class PHCFibersConical(Simulation):
    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        source=None,
        boundary_conditions=None,
        degree=(2, 2),
        pml_stretch=1 - 1j,
        beta=0,
        propagation_vector=(0, 0),
        eps=dolfin.DOLFIN_EPS,
        map_tol=1e-10,
        type="E",
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Lattice2D)
        if source is not None:
            assert source.dim == 2

        self.beta = beta
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)

        self.propagation_vector = propagation_vector

        Element1 = dolfin.FiniteElement("N1curl", geometry.mesh.ufl_cell(), degree[0])
        Element2 = dolfin.FiniteElement("CG", geometry.mesh.ufl_cell(), degree[1])
        element = dolfin.MixedElement([Element1, Element2])
        # Element3 = dolfin.FiniteElement("R", geometry.mesh.ufl_cell(), 0)
        # element = dolfin.MixedElement([Element1, Element2, Element3])

        self.periodic_bcs = BiPeriodic2D(geometry, map_tol=map_tol, eps=eps)
        function_space = ComplexFunctionSpace(
            geometry.mesh, element, constrained_domain=self.periodic_bcs
        )
        fs = dolfin.FunctionSpace(geometry.mesh, "DG", 0)
        el = fs.ufl_element()
        epsilon_coeff = Coefficient(self.epsilon, geometry, degree=0, element=el)
        mu_coeff = Coefficient(self.mu, geometry, degree=0, element=el)

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["box"]
        source_domains = []
        formulation = MaxwellConicalBands(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="box",
            modal=True,
            boundary_conditions=boundary_conditions,
            beta=beta,
            propagation_vector=propagation_vector,
            type=type,
        )
        super().__init__(geometry, formulation)

        self.degree = degree

    def eigensolve(self, *args, **kwargs):
        sol = super().eigensolve(*args, **kwargs)
        phasor = self.formulation.phasor
        self.solution["eigenvectors"] = [u * phasor for u in sol["eigenvectors"]]
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from pprint import pprint

import matplotlib.pyplot as plt
import numpy as np
import pytest
from scipy.spatial.transform import Rotation

from gyptis.utils import list_time

pytest_params = [("TM", 1), ("TE", 1), ("TM", 2), ("TE", 2)]

tol_balance = 1e-2
pmesh = 13


@pytest.mark.parametrize("polarization,degree", pytest_params)
def test_grating2d(polarization, degree):
    from gyptis import Grating, Layered
    from gyptis.models.grating2d import Grating2D, Layered2D, OrderedDict
    from gyptis.sources import PlaneWave

    wavelength = 1
    lmin = wavelength / pmesh
    period = 0.8
    R = period / 4

    thicknesses = OrderedDict(
        {
            "substrate": wavelength,
            "groove": 3 * R,
            "superstrate": wavelength,
        }
    )

    rot = Rotation.from_euler("zyx", [10, 0, 0], degrees=True)
    rmat = rot.as_matrix()
    eps_diff = rmat @ np.diag((8 - 0.1j, 3 - 0.1j, 4 - 0.2j)) @ rmat.T

    mu_groove = rmat @ np.diag((2 - 0.1j, 9 - 0.1j, 3 - 0.2j)) @ rmat.T

    epsilon = dict({"substrate": 1, "groove": 1, "superstrate": 1, "diff": eps_diff})
    mu = dict({"substrate": 1.6, "groove": mu_groove, "superstrate": 1, "diff": 1})

    geom = Layered(2, period, thicknesses, wavelength)
    yc = geom.y_position["groove"] + thicknesses["groove"] / 2
    diff = geom.add_circle(0, yc, 0, R)
    diff, groove = geom.fragment(diff, geom.layers["groove"])
    geom.add_physical(diff, "diff")
    geom.add_physical(groove, "groove")
    [geom.set_size(pml, lmin) for pml in geom.pml_physical]
    geom.set_size("groove", lmin)
    geom.set_size("diff", lmin)
    geom.set_size("superstrate", lmin)
    geom.set_size("substrate", lmin)
    geom.build()

    theta0 = 10
    angle = theta0 * np.pi / 180

    pw = PlaneWave(
        wavelength=wavelength, angle=angle, dim=2, domain=geom.mesh, degree=degree
    )
    s = Grating(geom, epsilon, mu, source=pw, degree=degree, polarization=polarization)

    s.solve()
    s.plot_field(field="diffracted")

    list_time()

    effs = s.diffraction_efficiencies(1, orders=True, subdomain_absorption=True)
    print(effs)

    if degree == 2:
        assert abs(effs["B"] - 1) < tol_balance, "Unsatified energy balance"

    s.plot_field(nper=3)
    s.plot_geometry(nper=3, c="k")

    # modal

    s = Grating2D(
        geom,
        epsilon,
        mu,
        modal=True,
        propagation_constant=0.6,
        degree=degree,
        polarization=polarization,
    )
    s.eigensolve(n_eig=6, target=6)
    s.solution["eigenvectors"]


@pytest.mark.parametrize("polarization,degree", pytest_params)
def test_grating2dpec(polarization, degree):
    from gyptis import Grating, Layered
    from gyptis.models.grating2d import Grating2D, Layered2D, OrderedDict
    from gyptis.sources import PlaneWave

    wavelength = 600
    period = 800
    h = 300
    w = 600
    theta0 = 30
    lmin = wavelength / pmesh
    lmin_pec = h / (pmesh * 2)
    thicknesses = OrderedDict(
        {
            "substrate": wavelength,
            "groove": wavelength,
            "superstrate": wavelength,
        }
    )
    rot = Rotation.from_euler("zyx", [20, 0, 0], degrees=True)
    rmat = rot.as_matrix()
    eps_groove = rmat @ np.diag((4 - 0.01j, 3 - 0.01j, 2 - 0.02j)) @ rmat.T
    mu_groove = rmat @ np.diag((2 - 0.001j, 4 - 0.001j, 3 - 0.002j)) @ rmat.T

    epsilon = dict({"substrate": 1.7, "groove": eps_groove, "superstrate": 1})
    mu = dict({"substrate": 1.2, "groove": mu_groove, "superstrate": 1})

    geom = Layered2D(period, thicknesses, wavelength)

    yc = geom.y_position["groove"] + thicknesses["groove"] / 2
    diff = geom.add_ellipse(0, yc, 0, w / 2, h / 2)
    groove = geom.cut(geom.layers["groove"], diff)

    geom.add_physical(groove, "groove")
    bnds = geom.get_boundaries("groove")
    geom.add_physical(bnds[-1], "hole", dim=1)

    for dom in ["substrate", "superstrate", "groove"] + geom.pml_physical:
        geom.set_size(dom, lmin)

    geom.set_size("hole", lmin_pec, dim=1)

    geom.build()

    angle = theta0 * np.pi / 180

    pw = PlaneWave(
        wavelength=wavelength, angle=angle, dim=2, domain=geom.mesh, degree=degree
    )

    boundary_conditions = {"hole": "PEC"}

    s = Grating2D(
        geom,
        epsilon,
        mu,
        source=pw,
        degree=degree,
        polarization=polarization,
        boundary_conditions=boundary_conditions,
    )
    s.solve()
    list_time()
    effs = s.diffraction_efficiencies(2, orders=True, subdomain_absorption=True)
    pprint(effs)
    if degree == 2:
        assert abs(effs["B"] - 1) < tol_balance, "Unsatified energy balance"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .simulation import *


class Waveguide2D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        wavenumber=0,
        boundary_conditions=None,
        degree=(2, 2),
        pml_stretch=1 - 1j,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, LayeredBoxPML2D)
        source = None

        self.wavenumber = wavenumber
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)

        Element1 = dolfin.FiniteElement("N1curl", geometry.mesh.ufl_cell(), degree[0])
        Element2 = dolfin.FiniteElement("CG", geometry.mesh.ufl_cell(), degree[1])
        element = dolfin.MixedElement([Element1, Element2])

        function_space = ComplexFunctionSpace(geometry.mesh, element)

        names = geometry.domains.keys()
        pmls_list = init_pmls(names, pml_stretch)
        degree_mat = max(degree)

        epsilon_coeff = Coefficient(
            self.epsilon, geometry, pmls=pmls_list, degree=degree_mat
        )
        mu_coeff = Coefficient(self.mu, geometry, pmls=pmls_list, degree=degree_mat)

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = pmls_list
        source_domains = []
        formulation = MaxwellWaveguide2D(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            boundary_conditions=boundary_conditions,
            wavenumber=wavenumber,
        )

        super().__init__(geometry, formulation)

        self.degree = max(degree)

    def solve(self, **kwargs):
        raise NotImplementedError

    def eigensolve(self, **kwargs):
        # kwargs["half"] = False
        self.solution = super().eigensolve(**kwargs)
        for i, (kz, v) in enumerate(
            zip(self.solution["eigenvalues"], self.solution["eigenvectors"])
        ):
            V0 = v[0] / Constant(kz)
            V1 = v[1] / Constant(kz)
            V2 = v[2] * Complex(0, 1)
            self.solution["eigenvectors"][i] = vector([V0, V1, V2])
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


def test_flags():
    import os
    import time

    import numpy as np

    import gyptis
    from gyptis import BoxPML, PlaneWave, Scattering
    from gyptis.utils import list_time

    gyptis.dolfin.set_log_level(6)

    # test = "-O3 -fno-align-functions -fno-align-jumps -fno-align-loops -fno-align-labels -march=native"
    # test = "-O3 -march=native"

    flags = ["-O0", "-O1", "-O2", "-O3", "-Ofast", "-Ofast -march=native"]
    flags = ["-O3"]
    tsolve = []
    tint = []

    polarization = "TM"
    degree = 2
    wavelength = 0.3
    pmesh = 3
    for cppflag in flags:
        print(f"-------   { cppflag }   --------")

        gyptis.dolfin.parameters["form_compiler"]["cpp_optimize_flags"] = cppflag

        lmin = wavelength / pmesh

        geom = BoxPML(
            dim=2,
            box_size=(4 * wavelength, 4 * wavelength),
            pml_width=(wavelength, wavelength),
        )
        cyl = geom.add_circle(0, 0, 0, 0.2)
        cyl, box = geom.fragment(cyl, geom.box)
        geom.add_physical(box, "box")
        geom.add_physical(cyl, "cyl")
        [geom.set_size(pml, lmin) for pml in geom.pmls]
        geom.set_size("box", lmin)
        geom.set_size("cyl", lmin)
        geom.build()

        epsilon = dict(box=1, cyl=3)
        mu = dict(box=1, cyl=1)
        pw = PlaneWave(
            wavelength=wavelength, angle=0, dim=2, domain=geom.mesh, degree=degree
        )

        for i in range(1, 2):
            s = Scattering(
                geom, epsilon, mu, pw, degree=degree, polarization=polarization
            )
            print(s.function_space.dim())
            t = -time.time()
            u = s.solve()
            t += time.time()
            print(f"elapsed time solve {t:0.3f}s")
            list_time()
            t1 = -time.time()
            gyptis.assemble(u * s.formulation.dx)
            t1 += time.time()
            print(f"elapsed time integral {t1:0.3f}s")
            # list_time()

            if i == 1:
                # skip the first as JIT compilation may happen
                tsolve.append(t)
                tint.append(t1)

    for cppflag, t, t1 in zip(flags, tsolve, tint):
        print(f"-------   { cppflag }   --------")
        print(f"elapsed time solve {t:0.3f}s")
        print(f"elapsed time integral {t1:0.3f}s")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import pytest

from gyptis import dolfin
from gyptis.complex import assemble, div, dot, grad, project
from gyptis.sources import *
from gyptis.utils.helpers import function2array, get_coordinates


def test_pw_2d():
    lambda0 = 0.1
    theta = np.pi / 6
    mesh = dolfin.UnitSquareMesh(50, 50)
    W = dolfin.FunctionSpace(mesh, "CG", 1)
    pw = plane_wave_2d(lambda0, theta, domain=mesh)
    uproj = project(pw, W)
    uarray = function2array(uproj.real) + 1j * function2array(uproj.imag)
    x, y = get_coordinates(W).T
    k0 = 2 * np.pi / lambda0
    kdotx = -k0 * (np.sin(theta) * x + np.cos(theta) * y)
    test = np.exp(1j * kdotx)
    err = abs(test - uarray) ** 2
    assert np.all(err < 1e-16)
    assert np.mean(err) < 1e-16


def test_pw_3d():
    theta, phi, psi = 1, 2, 3
    lambda0 = 0.1

    pw = plane_wave_3d(lambda0, theta, phi, psi)

    mesh = dolfin.UnitCubeMesh(10, 10, 10)
    W = dolfin.FunctionSpace(mesh, "CG", 1)

    x, y, z = get_coordinates(W).T
    k0 = 2 * np.pi / lambda0

    kx = -k0 * np.sin(theta) * np.cos(phi)
    ky = -k0 * np.sin(theta) * np.sin(phi)
    kz = -k0 * np.cos(theta)

    cx = np.cos(psi) * np.cos(theta) * np.cos(phi) - np.sin(psi) * np.sin(phi)
    cy = np.cos(psi) * np.cos(theta) * np.sin(phi) + np.sin(psi) * np.cos(phi)
    cz = -np.cos(psi) * np.sin(theta)
    kdotx = kx * x + ky * y + kz * z
    c = np.array([cx, cy, cz])
    for i in range(3):
        uproj = project(pw[i], W)
        pw_array = function2array(uproj.real) + 1j * function2array(uproj.imag)
        pw_test = np.exp(1j * kdotx) * c[i]
        err = abs(pw_test - pw_array) ** 2
        assert np.all(err < 1e-16)
        assert np.mean(err) < 1e-16


def test_gf_2d():
    mesh = dolfin.UnitSquareMesh(50, 50)
    degree = 1
    lambda0, xs, ys = 0.3, -0.1, -0.1
    GF = green_function_2d(lambda0, xs, ys, degree=degree, domain=mesh)
    k0 = 2 * np.pi / lambda0
    Helm = dot(grad(GF), grad(GF)) + k0**2 * GF * GF
    test = abs(assemble(Helm * dolfin.dx))
    print(test)
    assert test < 5e-3


def test_dipole():
    degree = 1
    mesh = dolfin.UnitSquareMesh(50, 50)
    Dipole(
        wavelength=0.1, position=(0.4, 0.2), angle=np.pi / 8, domain=mesh, degree=degree
    )


def test_gaussian_beam():
    degree = 1
    mesh = dolfin.UnitSquareMesh(50, 50)
    GaussianBeam(1, 0.4, 0.2, dim=2, domain=mesh, degree=degree)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


__all__ = ["tic", "toc", "list_time"]


import time

from .. import dolfin


def list_time():
    cols = [dolfin.TimingType.wall, dolfin.TimingType.system, dolfin.TimingType.user]
    return dolfin.list_timings(dolfin.TimingClear.clear, cols)


def tic():
    return time.time()


def toc(t0, verbose=True):
    t = time.time() - t0
    if verbose:
        print(f"elapsed time {t}s")
    return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .formulation import *

j = Complex(0, 1)
unit_vec_z = vector([0, 0, 1])


def curl_t(u):
    return (Dx(u[1], 0) - Dx(u[0], 1)) * unit_vec_z


def grad_t(u):
    du = grad(u)
    return vector([du[0], du[1], 0])


class MaxwellWaveguide2D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions=None,
        source_domains=None,
        reference=None,
        wavenumber=0,
        degree=1,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        if source_domains is None:
            source_domains = []
        super().__init__(
            geometry,
            coefficients,
            function_space,
            source=source,
            boundary_conditions=boundary_conditions,
            modal=True,
            degree=degree,
        )
        self.wavenumber = wavenumber
        self.source_domains = source_domains
        self.reference = reference
        self.epsilon, self.mu = self.coefficients
        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)
        self._mu_inv = self.mu.invert().as_property(dim=3)
        self._epsilon = self.epsilon.as_property(dim=3)
        self._epsilon_dom = self.epsilon.as_subdomain()
        self._mu_inv_dom = self.mu.invert().as_subdomain()

    def maxwell(self, u, v, domain="everywhere"):
        k0 = Constant(self.wavenumber)
        et, ez = vector([u[0], u[1], 0]), u[2]
        vt, vz = vector([v[0], v[1], 0]), v[2]
        if domain == []:
            return [0, 0] if self.modal else 0

        _eps = self._epsilon[domain]
        _mu_inv = self._mu_inv[domain]

        # _eps = self._epsilon_dom
        # _mu_inv = self._mu_inv_dom

        _ez = vector([0, 0, u[2]])
        _vz = vector([0, 0, v[2]])
        a_tt = inner(_mu_inv * curl_t(u), curl_t(v)) - (k0**2) * inner(_eps * et, vt)
        b_tt = inner(_mu_inv * et, vt)
        b_tz = inner(_mu_inv * et, grad_t(vz))
        b_zt = inner(_mu_inv * grad_t(ez), vt)
        b_zz = inner(_mu_inv * grad_t(ez), grad_t(vz)) - (k0**2) * inner(
            _eps * _ez, _vz
        )

        # a_tt = inner(_mu_inv * curl_t(u), curl_t(v)) - (k0**2) * inner(_eps * et, vt)
        # b_tt = inner(_mu_inv * et, vt)
        # b_tz = inner(_mu_inv * et, grad_t(vz))
        # b_zt = inner(_mu_inv * grad_t(ez), vt)
        # b_zz = inner(_mu_inv * grad_t(ez), grad_t(vz)) - (k0**2) * inner(
        #     _eps[2][2] * ez, vz
        # )

        f0 = a_tt
        f1 = b_tt + b_tz + b_zt + b_zz

        form = [f0, f1]
        if self.modal:
            return [form[0] * self.dx(domain), -form[1] * self.dx(domain)]
        raise NotImplementedError

    def _weak(self, u, v, u1):

        formulation = 0
        for dom in self.geometry.domains.keys():
            # for dom in ["everywhere"]:
            form = self.maxwell(u, v, domain=dom)
            if formulation == 0:
                formulation = form
            else:
                if self.modal:
                    for i in range(2):
                        formulation[i] += form[i]
                else:
                    formulation += form

        return (
            [f.real + f.imag for f in formulation]
            if self.modal
            else formulation.real + formulation.imag
        )

    @property
    def weak(self):
        u1 = Constant((0, 0, 0)) if self.modal else self.source.expression
        u = self.trial
        v = self.test
        return self._weak(u, v, u1)

    def build_pec_boundary_conditions(self, applied_function):
        if self.pec_boundaries != []:
            # FIXME: project is slow, avoid it.
            applied_function = project_iterative(
                applied_function, self.real_function_space
            )
            return build_pec_boundary_conditions(
                self.pec_boundaries,
                self.geometry,
                self.function_space,
                applied_function,
            )
        else:
            return []

    def build_boundary_conditions(self):
        applied_function = (
            Constant((0, 0, 0)) if self.modal else -self.source.expression
        )
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from .simulation import *


class Homogenization3D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        boundary_conditions=None,
        degree=1,
        direction="x",
        direct=True,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Lattice3D)

        self.geometry = geometry
        self.periodic_bcs = Periodic3D(geometry)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "CG", degree, constrained_domain=self.periodic_bcs
        )

        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)
        epsilon_coeff = Coefficient(self.epsilon, geometry, degree=degree, dim=3)
        mu_coeff = Coefficient(self.mu, geometry, degree=degree, dim=3)

        coefficients = epsilon_coeff, mu_coeff
        self.formulations = dict(epsilon={}, mu={})
        for case in ["epsilon", "mu"]:
            for direction in ["x", "y", "z"]:
                form = TwoScale3D(
                    geometry,
                    coefficients,
                    function_space,
                    degree=degree,
                    boundary_conditions=boundary_conditions,
                    direction=direction,
                    case=case,
                )
                self.formulations[case][direction] = form

        self.formulation = self.formulations["epsilon"]["x"]
        super().__init__(geometry, self.formulation, direct=direct)
        self.degree = degree
        self.solution = dict(epsilon={}, mu={})
        self.direct = direct
        self.direction = direction
        self.cell_volume = np.dot(
            np.cross(geometry.vectors[0], geometry.vectors[1]), geometry.vectors[2]
        )

    def solve_system(self, again=False):
        return super().solve_system(again=again, vector_function=False)

    def unit_cell_mean(self, f):
        return 1 / self.cell_volume * assemble(f * self.dx)

    def solve_param(self, case):
        self.formulation = self.formulations[case]["x"]
        super().__init__(self.geometry, self.formulation, direct=self.direct)
        phi_x = self.solve()
        self.formulations[case]["y"].build_rhs()
        self.vector = assemble(self.formulations[case]["y"].rhs)
        phi_y = self.solve_system(again=True)
        self.formulations[case]["z"].build_rhs()
        self.vector = assemble(self.formulations[case]["z"].rhs)
        phi_z = self.solve_system(again=True)
        self.solution[case] = dict(x=phi_x, y=phi_y, z=phi_z)
        return self.solution

    def _get_effective_param(self, case):
        self.solve_param(case)
        coeff = self.formulation.epsilon if case == "epsilon" else self.formulation.mu
        param = coeff.as_subdomain()
        if param.real.ufl_shape == (3, 3):
            param_mean = []
            for i in range(3):
                a = [self.unit_cell_mean(x) for x in param[i]]
                a = [_.real + 1j * _.imag for _ in a]
                param_mean.append(a)
            param_mean = np.array(param_mean)
        else:
            param_mean = self.unit_cell_mean(param)
            param_mean = param_mean.real + 1j * param_mean.imag
            param_mean *= np.eye(3)
        A = []
        for phi in self.solution[case].values():
            integrand = param * grad(phi)
            a = [self.unit_cell_mean(g) for g in integrand]
            a = [_.real + 1j * _.imag for _ in a]
            A.append(a)
        return param_mean - np.array(A)

    def get_effective_permittivity(self):
        return self._get_effective_param("epsilon")

    def get_effective_permeability(self):
        return self._get_effective_param("mu")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import os

import numpy as np
import pytest

import gyptis
from gyptis.utils import list_time

wavelength = 0.3
pmesh = 5
lmin = wavelength / pmesh


def build_geom():
    from gyptis import BoxPML

    geom = BoxPML(
        dim=2,
        box_size=(4 * wavelength, 4 * wavelength),
        pml_width=(wavelength, wavelength),
    )
    cyl = geom.add_circle(0, 0, 0, 0.2)
    cyl, box = geom.fragment(cyl, geom.box)
    geom.add_physical(box, "box")
    geom.add_physical(cyl, "cyl")
    [geom.set_size(pml, lmin) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.set_size("cyl", lmin)
    geom.build()
    return geom


@pytest.mark.parametrize(
    "degree,polarization", [(1, "TM"), (2, "TM"), (1, "TE"), (2, "TE")]
)
def test_scatt2d_pw(degree, polarization):
    from gyptis import PlaneWave, Scattering, dolfin

    epsilon = dict(box=1, cyl=3)
    mu = dict(box=1, cyl=1)

    geom = build_geom()
    mesh = geom.mesh

    pw = PlaneWave(wavelength=wavelength, angle=0, dim=2, domain=mesh, degree=degree)

    s = Scattering(geom, epsilon, mu, pw, degree=degree, polarization=polarization)
    u = s.solve()
    list_time()
    print(gyptis.assemble(u * s.formulation.dx))
    if gyptis.ADJOINT:
        from gyptis import project

        eps_max, eps_min = 3, 1
        Actrl = dolfin.FunctionSpace(mesh, "DG", 0)
        ctrl0 = dolfin.Expression("0.1", degree=2)
        ctrl = project(ctrl0, Actrl)
        eps_lens_func = ctrl * (eps_max - eps_min) + eps_min
        eps_lens_func *= gyptis.Complex(1, 0)
        dolfin.set_working_tape(dolfin.Tape())
        h = dolfin.Function(Actrl)
        h.vector()[:] = 1e-2 * np.random.rand(Actrl.dim())
        # epsilon["cyl"] = project(eps_lens_func,s.formulation.real_function_space)
        epsilon["cyl"] = eps_lens_func
        # project(eps_lens_func,s.formulation.real_function_space)
        s = Scattering(geom, epsilon, mu, pw, degree=degree, polarization=polarization)
        field = s.solve()
        J = -gyptis.assemble(gyptis.inner(field, field.conj) * s.dx("box")).real
        Jhat = dolfin.ReducedFunctional(J, dolfin.Control(ctrl))
        conv_rate = dolfin.taylor_test(Jhat, ctrl, h)
        print("convergence rate = ", conv_rate)
        assert abs(conv_rate - 2) < 1e-2


@pytest.mark.parametrize(
    "degree,polarization", [(1, "TM"), (2, "TM"), (1, "TE"), (2, "TE")]
)
def test_scatt2d_ls(degree, polarization):
    from gyptis import LineSource, Scattering

    epsilon = dict(box=1, cyl=3)
    mu = dict(box=1, cyl=1)

    geom = build_geom()
    mesh = geom.mesh
    gf = LineSource(wavelength, (-wavelength, 0), domain=mesh, degree=degree)
    s = Scattering(geom, epsilon, mu, gf, degree=degree, polarization=polarization)
    u = s.solve()
    list_time()
    gf.expression + u
    s.source.position = (0, -wavelength)
    s.assemble_rhs()
    s.solve_system(again=True)
    list_time()

    s.plot_field()
    s.animate_field(n=2, filename="animation.gif")
    os.system("rm -rf animation.gif")


@pytest.mark.parametrize("polarization", ["TM", "TE"])
def test_scatt2d_pec(polarization):
    from gyptis import BoxPML, PlaneWave, Scattering

    geom = BoxPML(
        dim=2,
        box_size=(4 * wavelength, 4 * wavelength),
        pml_width=(wavelength, wavelength),
    )
    cyl = geom.add_circle(0, 0, 0, 0.2)
    box = geom.cut(geom.box, cyl)
    geom.add_physical(box, "box")
    cyl_bnds = geom.get_boundaries("box")[-1]
    geom.add_physical(cyl_bnds, "cyl_bnds", dim=1)
    [geom.set_size(pml, lmin) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.build()
    mesh = geom.mesh_object["mesh"]
    epsilon = dict(box=1)
    mu = dict(box=1)

    pw = PlaneWave(wavelength=wavelength, angle=0, dim=2, domain=mesh, degree=2)

    bcs = {"cyl_bnds": "PEC"}
    s = Scattering(
        geom,
        epsilon,
        mu,
        pw,
        degree=2,
        polarization=polarization,
        boundary_conditions=bcs,
    )

    u = s.solve()
    list_time()
    print(gyptis.assemble(u * s.formulation.dx))


@pytest.mark.parametrize("polarization", ["TM", "TE"])
def test_scatt2d_scs(polarization):
    from gyptis import BoxPML, PlaneWave, Scattering

    pmesh = 10
    wavelength = 452
    eps_core = 2
    eps_shell = 6

    R1 = 60
    R2 = 30
    Rcalc = 2 * R1
    lmin = wavelength / pmesh
    pml_width = wavelength

    lbox = Rcalc * 2 * 1.1
    geom = BoxPML(
        dim=2,
        box_size=(lbox, lbox),
        pml_width=(pml_width, pml_width),
        Rcalc=Rcalc,
    )
    box = geom.box
    shell = geom.add_circle(0, 0, 0, R1)
    out = geom.fragment(shell, box)
    box = out[1:3]
    shell = out[0]
    core = geom.add_circle(0, 0, 0, R2)
    core, shell = geom.fragment(core, shell)
    geom.add_physical(box, "box")
    geom.add_physical(core, "core")
    geom.add_physical(shell, "shell")
    [geom.set_size(pml, lmin * 0.7) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.set_size("core", lmin / eps_core**0.5)
    geom.set_size("shell", lmin / eps_shell**0.5)
    geom.build()
    pw = PlaneWave(wavelength=wavelength, angle=0, dim=2, domain=geom.mesh, degree=2)
    epsilon = dict(box=1, core=eps_core, shell=eps_shell)
    mu = dict(box=1, core=1, shell=1)

    s = Scattering(
        geom,
        epsilon,
        mu,
        pw,
        degree=2,
        polarization=polarization,
    )
    s.solve()
    cs = s.get_cross_sections()
    assert np.allclose(
        cs["extinction"], cs["scattering"] + cs["absorption"], rtol=1e-11
    )
    print(cs["extinction"])
    print(cs["scattering"] + cs["absorption"])
    print(abs(cs["scattering"] + cs["absorption"] - cs["extinction"]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell2d import *


class Maxwell2DBands(Maxwell2D):
    def __init__(self, *args, propagation_vector=(0, 0), modal=True, **kwargs):
        super().__init__(*args, **kwargs, modal=modal)
        self.propagation_vector = propagation_vector

    @property
    def phasor(self):
        _phasor = phasor(
            self.propagation_vector[0],
            direction=0,
            degree=self.degree,
            domain=self.geometry.mesh,
        )
        _phasor *= phasor(
            self.propagation_vector[1],
            direction=1,
            degree=self.degree,
            domain=self.geometry.mesh,
        )
        return _phasor

    @property
    def weak(self):
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return super()._weak(u, v, Constant(0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import pytest

from gyptis.utils.parallel import *


def f(x, y, a=1, b="b"):
    print(f"f called with {x} {y} {a} {b}")
    y = x**2
    print(f"output y = {x}^2 = {y}")
    return y


@parloop(n_jobs=4)
def fpar(x, *args, **kwargs):
    return f(x, *args, **kwargs)


def test_para():
    x = [1, 2, 3, 4]
    y = [f(_, 9) for _ in x]
    ypar = fpar(x, 9)
    assert ypar == y


def test_mpi_print():
    mpi_print("Hello world!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from IPython import get_ipython

from .helpers import *
from .jupyter import VersionTable
from .log import *
from .parallel import *
from .sample import *
from .time import *

_IP = get_ipython()
if _IP is not None:
    _IP.register_magics(VersionTable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io
"""
Information about the package.
"""

import importlib.metadata as metadata
from email.utils import parseaddr

try:
    data = metadata.metadata("gyptis")
    __version__ = metadata.version("gyptis")
    __author__, _ = parseaddr(data.get("Author-Email"))
    __description__ = data.get("summary")
except Exception:
    __version__ = "unknown"
    __author__ = "unknown"
    __description__ = "unknown"

__name__ = "gyptis"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from collections import OrderedDict

import numpy as np
from numpy.linalg import inv
from scipy.constants import c, epsilon_0, mu_0, pi

from ..materials import Subdomain, complex_vector
from .source import *


def field_stack_2D(phi, alpha, beta, yshift=0, degree=1, domain=None):
    """
    Compute the electric field associated with a stack of 2D layers.

    Parameters
    ----------
    phi : complex of shape (2,)
        The coefficients of the + and - going waves.
    alpha : complex
        The x component of the wavevector.
    beta : complex
        The y component of the wavevector.
    yshift : float, optional
        The shift of the stack along the y-axis. Default is 0.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : Optional[Domain]
        The domain of the output Expression. Default is None.

    Returns
    -------
    expr : Expression
        The electric field associated with the stack of 2D layers.
    """
    alpha0_re, alpha0_im, beta0_re, beta0_im = sp.symbols(
        "alpha0_re,alpha0_im,beta0_re,beta0_im", real=True
    )
    alpha0 = alpha0_re + 1j * alpha0_im
    beta0 = beta0_re + 1j * beta0_im
    Kplus = sympyvector((alpha0, beta0, 0))
    Kminus = sympyvector((alpha0, -beta0, 0))
    deltaY = sympyvector((0, sp.symbols("yshift", real=True), 0))

    def pw(K):
        return sp.exp(1j * K.dot(X - deltaY))

    phi_plus_re, phi_plus_im = sp.symbols("phi_plus_re,phi_plus_im", real=True)
    phi_minus_re, phi_minus_im = sp.symbols("phi_minus_re,phi_minus_im", real=True)
    phi_plus = phi_plus_re + 1j * phi_plus_im
    phi_minus = phi_minus_re + 1j * phi_minus_im
    field = phi_plus * pw(Kplus) + phi_minus * pw(Kminus)

    expr = expression2complex_2d(
        field,
        alpha0_re=alpha.real,
        alpha0_im=alpha.imag,
        beta0_re=beta.real,
        beta0_im=beta.imag,
        phi_plus_re=phi[0].real,
        phi_plus_im=phi[0].imag,
        phi_minus_re=phi[1].real,
        phi_minus_im=phi[1].imag,
        yshift=yshift,
        degree=degree,
        domain=domain,
    )
    return expr


def field_stack_3D(phi, alpha, beta, gamma, zshift=0, degree=1, domain=None):
    """
    Compute the electric field associated with a stack of 3D layers.

    Parameters
    ----------
    phi : list of complex
        The coefficients of the + and - going waves for each component (x, y, z).
    alpha : complex
        The x component of the wavevector.
    beta : complex
        The y component of the wavevector.
    gamma : complex
        The z component of the wavevector.
    zshift : float, optional
        The shift of the stack along the z-axis. Default is 0.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : Optional[Domain]
        The domain of the output Expression. Default is None.

    Returns
    -------
    Complex
        The electric field associated with the stack of 3D layers expressed
        as a complex tensor.
    """

    alpha0_re, alpha0_im, beta0_re, beta0_im, gamma0_re, gamma0_im = sp.symbols(
        "alpha0_re, alpha0_im, beta0_re, beta0_im, gamma0_re, gamma0_im", real=True
    )
    alpha0 = alpha0_re + 1j * alpha0_im
    beta0 = beta0_re + 1j * beta0_im
    gamma0 = gamma0_re + 1j * gamma0_im
    Kplus = sympyvector((alpha0, beta0, gamma0))
    Kminus = sympyvector((alpha0, beta0, -gamma0))
    deltaZ = sympyvector((0, 0, sp.symbols("zshift", real=True)))

    def pw(K):
        return sp.exp(1j * K.dot(X - deltaZ))

    fields = []
    for comp in ["x", "y", "z"]:
        phi_plus_re, phi_plus_im = sp.symbols(
            f"phi_plus_{comp}_re,phi_plus_{comp}_im", real=True
        )
        phi_minus_re, phi_minus_im = sp.symbols(
            f"phi_minus_{comp}_re,phi_minus_{comp}_im", real=True
        )
        phi_plus = phi_plus_re + 1j * phi_plus_im
        phi_minus = phi_minus_re + 1j * phi_minus_im
        field = phi_plus * pw(Kplus) + phi_minus * pw(Kminus)
        fields.append(field)
    code = [[sp.printing.ccode(p) for p in f.as_real_imag()] for f in fields]
    code = np.ravel(code).tolist()
    expr = [
        dolfin.Expression(
            c,
            alpha0_re=alpha.real,
            alpha0_im=alpha.imag,
            beta0_re=beta.real,
            beta0_im=beta.imag,
            gamma0_re=gamma.real,
            gamma0_im=gamma.imag,
            phi_plus_x_re=phi[0].real,
            phi_plus_x_im=phi[0].imag,
            phi_minus_x_re=phi[1].real,
            phi_minus_x_im=phi[1].imag,
            phi_plus_y_re=phi[2].real,
            phi_plus_y_im=phi[2].imag,
            phi_minus_y_re=phi[3].real,
            phi_minus_y_im=phi[3].imag,
            phi_plus_z_re=phi[4].real,
            phi_plus_z_im=phi[4].imag,
            phi_minus_z_re=phi[5].real,
            phi_minus_z_im=phi[5].imag,
            zshift=zshift,
            degree=degree,
            domain=domain,
        )
        for c in code
    ]

    return Complex(
        as_tensor([expr[0], expr[2], expr[4]]), as_tensor([expr[1], expr[3], expr[5]])
    )


def solve(thicknesses, eps, mu, lambda0, theta0, phi0, psi0):
    """
    Solve the electromagnetic field transmission and reflection through a stack of layers.

    Parameters
    ----------
    thicknesses : list of float
        The thickness of each layer in the stack.
    eps : list of complex
        The permittivity of each layer in the stack.
    mu : list of complex
        The permeability of each layer in the stack.
    lambda0 : float
        The wavelength of the incident wave.
    theta0 : float
        The incident angle of the wave in radians.
    phi0 : float
        The azimuthal angle of the wave in radians.
    psi0 : float
        The polarization angle of the wave in radians.

    Returns
    -------
    phi : list of ndarray
        The field coefficients for each layer in the stack.
    propagation_constants : tuple
        The propagation constants (alpha0, beta0, gamma) for the stack.
    efficiencies : dict
        A dictionary containing the reflection (R), transmission (T), and
        absorption (Q) efficiencies.
    """

    k0 = 2 * pi / lambda0
    omega = k0 * c

    alpha0 = -k0 * np.sin(theta0) * np.cos(phi0)
    beta0 = -k0 * np.sin(theta0) * np.sin(phi0)
    # gamma0 = -k0 * np.cos(theta0)
    Ex0 = np.cos(psi0) * np.cos(theta0) * np.cos(phi0) - np.sin(psi0) * np.sin(phi0)
    Ey0 = np.cos(psi0) * np.cos(theta0) * np.sin(phi0) + np.sin(psi0) * np.cos(phi0)
    # Ez0 = -np.cos(psi0) * np.sin(theta0)

    def _matrix_pi(M, gamma):
        q = gamma * M
        return np.array(
            [
                [1, 1, 0, 0],
                [0, 0, 1, 1],
                [q[0, 1], -q[0, 1], -q[0, 0], q[0, 0]],
                [q[1, 1], -q[1, 1], -q[1, 0], q[1, 0]],
            ],
            dtype=complex,
        )

    def _matrix_t(gamma, e):
        t = np.zeros((4, 4), dtype=complex)
        t[0, 0] = t[2, 2] = np.exp(1j * gamma * e)
        t[1, 1] = t[3, 3] = np.exp(-1j * gamma * e)
        return t

    def _matrix_B(eps, mu):
        eps *= epsilon_0
        mu *= mu_0
        return np.array(
            [
                [omega * mu, 0, beta0],
                [0, omega * mu, -alpha0],
                [-beta0, alpha0, -omega * eps],
            ]
        )

    gamma = [np.sqrt(k0**2 * e * m - alpha0**2 - beta0**2) for e, m in zip(eps, mu)]
    B = [_matrix_B(e, m) for e, m in zip(eps, mu)]
    M = [inv(b) for b in B]
    Pi = [_matrix_pi(m, g) for m, g in zip(M, gamma)]
    T_ = [_matrix_t(g, e) for g, e in zip(gamma[1:-1], thicknesses)]
    Tr = T_ + [np.eye(4)]

    M1 = np.eye(4)
    p_prev = Pi[0]
    for p, t in zip(Pi[1:], Tr):
        M1 = inv(t) @ inv(p) @ p_prev @ M1
        p_prev = p

    K = inv(M1)
    Q = np.array(
        [
            [K[0, 0], 0, K[0, 2], 0],
            [K[1, 0], -1, K[1, 2], 0],
            [K[2, 0], 0, K[2, 2], 0],
            [K[3, 0], 0, K[3, 2], -1],
        ],
        dtype=complex,
    )

    # solve
    U0 = np.array([Ex0, 0, Ey0, 0], dtype=complex)
    sol = inv(Q) @ U0

    # get coefficients by recurrence
    phi_0 = np.array([Ex0, sol[1], Ey0, sol[3]])
    # phi_end = np.array([sol[0], 0, sol[2], 0])
    p_prev = Pi[0]
    phi_prev = phi_0
    phi = [phi_0]

    for p, t in zip(Pi[1:], Tr):
        phi_j = inv(t) @ inv(p) @ p_prev @ phi_prev
        phi.append(phi_j)
        p_prev = p
        phi_prev = phi_j

    # assert np.all(np.abs(phi_j - phi_end)<1e-14)

    # Ez
    for i, (p, g, m) in enumerate(zip(phi.copy(), gamma, M)):
        phiz_plus = (m[2, 0] * p[2] - m[2, 1] * p[0]) * g
        phiz_minus = (m[2, 1] * p[1] - m[2, 0] * p[3]) * g
        phixh_plus = (m[0, 0] * p[2] - m[0, 1] * p[0]) * g
        phixh_minus = (m[0, 1] * p[1] - m[0, 0] * p[3]) * g
        phiyh_plus = (m[1, 0] * p[2] - m[1, 1] * p[0]) * g
        phiyh_minus = (m[1, 1] * p[1] - m[1, 0] * p[3]) * g
        phizh_plus = (m[2, 1] * phixh_plus - m[2, 0] * phiyh_plus) * g
        phizh_minus = (m[2, 0] * phiyh_minus - m[2, 1] * phixh_minus) * g
        phi[i] = np.append(
            phi[i],
            [
                phiz_plus,
                phiz_minus,
                phixh_plus,
                phixh_minus,
                phiyh_plus,
                phiyh_minus,
                phizh_plus,
                phizh_minus,
            ],
        )

    R = (
        1.0
        / gamma[0] ** 2
        * (
            (gamma[0] ** 2 + alpha0**2) * abs(phi[0][1]) ** 2
            + (gamma[0] ** 2 + beta0**2) * abs(phi[0][3]) ** 2
            + 2 * alpha0 * beta0 * np.real(phi[0][1] * phi[0][3].conjugate())
        )
    )
    T = (
        1.0
        / (gamma[0] * gamma[-1] * mu[-1])
        * (
            (gamma[-1] ** 2 + alpha0**2) * abs(phi[-1][0]) ** 2
            + (gamma[-1] ** 2 + beta0**2) * abs(phi[-1][2]) ** 2
            + 2 * alpha0 * beta0 * np.real(phi[-1][0] * phi[-1][2].conjugate())
        )
    )

    losses = np.array(eps).imag

    P0 = 0.5 * np.sqrt(epsilon_0 / mu_0) * np.cos(theta0)

    th = [0] + thicknesses + [0]
    tcum = np.cumsum([0] + thicknesses).tolist()
    tcum += [tcum[-1]]

    q = [
        loss * e * sum(abs(p) ** 2)
        + 2
        * loss
        * np.real(
            p[0]
            * p[1].conj()
            # * np.exp(-2j * g * zi)
            * (np.exp(-2j * g * zi) - 1)
            / (2j * g)
        )
        for loss, p, e, zi, g in zip(losses, phi, th, tcum, gamma)
    ]

    Q_ = [-0.5 * epsilon_0 * omega / P0 * q_ for q_ in q]
    Q = sum(Q_)
    propagation_constants = alpha0, beta0, gamma
    efficiencies = dict(R=R, T=T, Q=Q)

    return phi, propagation_constants, efficiencies


def get_coeffs_stack(config, lambda0, theta0, phi0, psi0):
    """
    Compute the coefficients of the stack of layers.

    Parameters
    ----------
    config : OrderedDict
        A dictionary containing the configuration of the stack.
    lambda0 : float
        The wavelength of the incident wave.
    theta0 : float
        The angle of incidence.
    phi0 : float
        The angle of the electric field.
    psi0 : float
        The angle of the magnetic field.

    Returns
    -------
    coeffs : list of 4 arrays
        The coefficients of the stack of layers.
    propagation_constants : tuple of 3 floats
        The propagation constants of the stack of layers.
    efficiencies : dict
        The efficiencies of the stack of layers.
    """
    eps = [d["epsilon"] for d in config.values()]
    mu = [d["mu"] for d in config.values()]
    thicknesses = [d["thickness"] for d in config.values() if "thickness" in d.keys()]
    return solve(thicknesses, eps, mu, lambda0, theta0, phi0, psi0)


def get_stack_efficiencies(thicknesses, epsilon, mu, lambda0, angles):
    """
    Compute the efficiencies of the stack of layers.

    Parameters
    ----------
    thicknesses : list of float
        The thicknesses of the layers.
    epsilon : Complex
        The relative permittivity of each layer.
    mu : Complex
        The relative permeability of each layer.
    lambda0 : float
        The wavelength of the incident wave.
    angles : tuple of 3 floats
        The angles of incidence, electric field and magnetic field.

    Returns
    -------
    efficiencies : dict
        The efficiencies of the stack of layers.
    """
    out = solve(thicknesses, epsilon, mu, lambda0, *angles)
    return out[-1]


def make_stack(
    geometry,
    coefficients,
    plane_wave,
    polarization="TM",
    source_domains=None,
    degree=1,
    dim=2,
):
    """
    Build a stack of layers.

    Parameters
    ----------
    geometry : Geometry
        The geometry of the problem.
    coefficients : tuple of Coefficients
        The coefficients of the materials.
    plane_wave : PlaneWave
        The incident plane wave.
    polarization : str, optional
        The polarization of the incident wave. Default is "TM".
    source_domains : list of str, optional
        The domains where the source is applied. Default is an empty list.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    dim : int, optional
        The dimension of the problem. Default is 2.

    Returns
    -------
    annex_field : dict
        A dictionary containing the incident and stack fields.
    """
    if source_domains is None:
        source_domains = []
    epsilon, mu = coefficients
    config = OrderedDict(
        {
            "superstrate": {
                "epsilon": epsilon.dict["superstrate"],
                "mu": mu.dict["superstrate"],
            },
            "substrate": {
                "epsilon": epsilon.dict["substrate"],
                "mu": mu.dict["substrate"],
            },
        }
    )
    inc_field = {}
    stack_field = {}
    if dim == 2:
        if polarization == "TM":
            _psi = np.pi / 2
            _phi_ind = 2
        else:
            _psi = 0
            _phi_ind = 8

        stack_output = get_coeffs_stack(
            config,
            plane_wave.wavelength,
            plane_wave.angle,
            0,
            _psi,
        )
        phi_, propagation_constants, efficiencies_stack = stack_output
        (
            alpha0,
            _,
            beta,
        ) = propagation_constants
        phi = [[p[_phi_ind], p[_phi_ind + 1]] for p in phi_]
        phi = (np.array(phi) / phi[0][0]).tolist()

        u_stack = [
            field_stack_2D(p, alpha0, b, yshift=0, domain=geometry.mesh)
            for p, b in zip(phi, beta)
        ]
        phi0 = [phi[0][0], 0]
        u_0 = field_stack_2D(phi0, alpha0, beta[0], yshift=0, domain=geometry.mesh)
        estack = dict(zip(config.keys(), u_stack))

        for dom in source_domains:
            estack[dom] = estack["superstrate"]

        e0 = {"superstrate": u_0}
        for dom in source_domains:
            e0[dom] = e0["superstrate"]
        e0["substrate"] = Complex(0, 0)

        for dom in geometry.pml_physical:
            estack[dom] = e0[dom] = Complex(0, 0)

        ustack_coeff = Subdomain(
            geometry.markers,
            geometry.domains,
            estack,
            degree=degree,
            domain=geometry.mesh,
        )
        u0_coeff = Subdomain(
            geometry.markers, geometry.domains, e0, degree=degree, domain=geometry.mesh
        )

        for dom in epsilon.dict.keys():
            inc_field[dom] = e0[dom]
            stack_field[dom] = estack[dom]
    else:
        stack_output = get_coeffs_stack(
            config,
            plane_wave.wavelength,
            plane_wave.angle[0],
            plane_wave.angle[1],
            plane_wave.angle[2],
        )
        phi_, propagation_constants, efficiencies_stack = stack_output
        (
            alpha0,
            beta0,
            gamma,
        ) = propagation_constants
        phi = [p[:6] for p in phi_]

        u_stack = [
            field_stack_3D(p, alpha0, beta0, g, domain=geometry.mesh)
            for p, g in zip(phi, gamma)
        ]
        phi0 = np.zeros_like(phi[0])
        phi0[::2] = phi[0][::2]
        u_0 = field_stack_3D(phi0, alpha0, beta0, gamma[0], domain=geometry.mesh)

        ustack_coeff = []
        u0_coeff = []
        estack_list = []
        e0_list = []
        for comp in range(3):
            estack = {k: v[comp] for k, v in zip(config.keys(), u_stack)}
            for dom in source_domains:
                estack[dom] = estack["superstrate"]

            e0 = {"superstrate": u_0}
            for dom in source_domains:
                e0[dom] = e0["superstrate"]
            e0["substrate"] = Complex(0, 0)

            for dom in geometry.pml_physical:
                estack[dom] = e0[dom] = Complex(0, 0)

            _ustack_coeff = Subdomain(
                geometry.markers,
                geometry.domains,
                estack,
                degree=degree,
                domain=geometry.mesh,
            )
            _u0_coeff = Subdomain(
                geometry.markers,
                geometry.domains,
                e0,
                degree=degree,
                domain=geometry.mesh,
            )
            ustack_coeff.append(_ustack_coeff)
            u0_coeff.append(_u0_coeff)
            estack_list.append(estack)
            e0_list.append(e0)

        ustack_coeff = complex_vector(ustack_coeff)
        u0_coeff = complex_vector(u0_coeff)

        for dom in epsilon.dict.keys():
            inc_field[dom] = complex_vector([e0[dom] for e0 in e0_list])
            stack_field[dom] = complex_vector([estack[dom] for estack in estack_list])

    annex_field_dict = {"incident": inc_field, "stack": stack_field}
    annex_field_coeff = {"incident": u0_coeff, "stack": ustack_coeff}
    annex_field = dict(as_subdomain=annex_field_coeff, as_dict=annex_field_dict)
    annex_field["stack_output"] = stack_output
    annex_field["phi"] = phi
    return annex_field



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


"""
Scattering Matrix calculations for a scatterer.

See:
Demésy, G., Auger, J.-C. & Stout, B. Scattering matrix of arbitrarily shaped
objects: combining finite elements and vector partial waves.
J. Opt. Soc. Am. A, JOSAA 35, 1401–1409 (2018)
"""

import scipy.special as sp

from .. import dolfin as df
from .. import pi
from ..complex import Complex, Constant, j, vector
from .source import Source


def sqrt_safe(x):
    """
    Compute a safe square root of a complex number.

    This function computes the square root of a complex number while avoiding
    the branch cut of the complex square root. The output is a complex number
    with positive real part.

    Parameters
    ----------
    x : Complex
        The complex number to compute the square root of.

    Returns
    -------
    Complex
        The square root of x with positive real part.
    """
    x = Complex(x, 0)
    t = x.phase
    return df.sqrt(x.module) * Complex(df.cos(t / 2), df.sin(t / 2))


def sph_bessel_J(nu, f, degree=1, domain=None):
    """
    Compute the spherical Bessel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Bessel function.
    f : float
        The argument of the spherical Bessel function.

    Returns
    -------
    Complex
        The spherical Bessel function of the first kind of order nu.
    """

    return df.Expression(
        "sqrt(pi / (2 * f)) * jn(nu + 0.5, f)", nu=nu, f=f, degree=degree, domain=domain
    )


def sph_bessel_Y(nu, f):
    """
    Compute the spherical Bessel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Bessel function.
    f : float
        The argument of the spherical Bessel function.

    Returns
    -------
    Complex
        The spherical Bessel function of the second kind of order nu.
    """
    return df.Expression(
        "sqrt(pi / (2 * f)) * yn(nu + 0.5, f)", nu=nu, f=f, degree=degree, domain=domain
    )


def sph_bessel_H1(nu, f):
    """
    Compute the spherical Hankel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Hankel function.
    f : float
        The argument of the spherical Hankel function.

    Returns
    -------
    Complex
        The spherical Hankel function of the first kind of order nu.
    """
    return Complex(sph_bessel_J(nu, f), sph_bessel_Y(nu, f))


def sph_bessel_H2(nu, f):
    """
    Compute the spherical Hankel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Hankel function.
    f : float
        The argument of the spherical Hankel function.

    Returns
    -------
    Complex
        The spherical Hankel function of the second kind of order nu.
    """
    return Complex(sph_bessel_J(nu, f), -sph_bessel_Y(nu, f))


def ricatti_bessel_J(nu, f):
    """
    Compute the Ricatti-Bessel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the first kind of order nu.
    """

    return f * sph_bessel_J(nu, f)


def ricatti_bessel_Y(nu, f):
    """
    Compute the Ricatti-Bessel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the second kind of order nu.
    """

    return f * sph_bessel_Y(nu, f)


def ricatti_bessel_H1(nu, f):
    """
    Compute the Ricatti-Bessel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the first kind of order nu.
    """

    return Complex(ricatti_bessel_J(nu, f), ricatti_bessel_Y(nu, f))


def ricatti_bessel_H2(nu, f):
    """
    Compute the Ricatti-Bessel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the second kind of order nu.
    """
    return Complex(ricatti_bessel_J(nu, f), -ricatti_bessel_Y(nu, f))


def sph_jn(nu, f):
    """
    Compute the spherical Bessel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Bessel function.
    f : float
        The argument of the spherical Bessel function.

    Returns
    -------
    float
        The spherical Bessel function of the first kind of order nu.
    """
    return (pi / (2 * f)) ** 0.5 * sp.jv(nu + 0.5, f)


def sph_yn(nu, f):
    """
    Compute the spherical Bessel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Bessel function.
    f : float
        The argument of the spherical Bessel function.

    Returns
    -------
    float
        The spherical Bessel function of the second kind of order nu.
    """

    return (pi / (2 * f)) ** 0.5 * sp.yv(nu + 0.5, f)


def sph_hn1(nu, f):
    """
    Compute the spherical Hankel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Hankel function.
    f : float
        The argument of the spherical Hankel function.

    Returns
    -------
    Complex
        The spherical Hankel function of the first kind of order nu.
    """
    return sph_jn(nu, f) + 1j * sph_yn(nu, f)


def sph_hn2(nu, f):
    """
    Compute the spherical Hankel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the spherical Hankel function.
    f : float
        The argument of the spherical Hankel function.

    Returns
    -------
    Complex
        The spherical Hankel function of the second kind of order nu.
    """
    return sph_jn(nu, f) - 1j * sph_yn(nu, f)


def rb_jn(nu, f):
    """
    Compute the Ricatti-Bessel function of the first kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the first kind of order nu.
    """
    return f * sph_jn(nu, f)


def rb_yn(nu, f):
    """
    Compute the Ricatti-Bessel function of the second kind of order nu.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the second kind of order nu.
    """
    return f * sph_yn(nu, f)


def rb_hn1(nu, f):
    """
    Compute the Ricatti-Bessel function of the first kind of order nu,
    using the spherical Bessel functions of the first and second kinds.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the first kind of order nu.
    """

    return rb_jn(nu, f) + 1j * rb_yn(nu, f)


def rb_hn2(nu, f):
    """
    Compute the Ricatti-Bessel function of the second kind of order nu,
    using the spherical Bessel functions of the first and second kinds.

    Parameters
    ----------
    nu : float
        The order of the Ricatti-Bessel function.
    f : float
        The argument of the Ricatti-Bessel function.

    Returns
    -------
    Complex
        The Ricatti-Bessel function of the second kind of order nu.
    """

    return rb_jn(nu, f) - 1j * rb_yn(nu, f)


def nm2p(n, m):
    """
    Convert a pair of spherical harmonic indices (n, m) to a scalar index p.

    Parameters
    ----------
    n : int
        The degree of the spherical harmonic.
    m : int
        The order of the spherical harmonic.

    Returns
    -------
    int
        The scalar index of the spherical harmonic.
    """
    return n * (n + 1) - m


def p2nm(p):
    """
    Convert a scalar index p of a spherical harmonic to a pair of spherical
    harmonic indices (n, m).

    Parameters
    ----------
    p : int
        The scalar index of the spherical harmonic.

    Returns
    -------
    tuple
        The pair of spherical harmonic indices (n, m).
    """
    n = int(p**0.5)
    m = n * (n + 1) - p
    return n, m


# Recurrence relations for scalar spherical harmonics


def P(x, n, m, y=None):
    """
    Compute the associated Legendre polynomial of degree n and order m,
    evaluated at x.

    Parameters
    ----------
    x : float
        The argument of the associated Legendre polynomial.
    n : int
        The degree of the associated Legendre polynomial.
    m : int
        The order of the associated Legendre polynomial.
    y : float, optional
        The value of sqrt(1 - x**2). If not provided, it is computed.

    Returns
    -------
    float
        The associated Legendre polynomial of degree n and order m, evaluated
        at x.
    """
    y = y or sqrt_safe(1 - x**2)
    if m < 0:
        return (-1) ** (-m) * P(x, n, -m)
    elif n == m == 0:
        return 1 / (4 * pi) ** 0.5
    elif n == m:
        return -(((2 * n + 1) / (2 * n)) ** 0.5) * y * P(x, n - 1, m - 1)
    elif m == (n - 1):
        return x * (2 * n + 1) ** 0.5 * P(x, n - 1, n - 1)
    else:
        return (((2 * n + 1) / (n**2 - m**2)) ** 0.5) * (
            (2 * n - 1) ** 0.5 * x * P(x, n - 1, m)
            - (((n - 1) ** 2 - m**2) / (2 * n - 3)) ** 0.5 * P(x, n - 2, m)
        )


def u(x, n, m, y=None):
    """
    Compute the derivative of the associated Legendre polynomial of degree n and
    order m, evaluated at x.

    Parameters
    ----------
    x : float
        The argument of the associated Legendre polynomial.
    n : int
        The degree of the associated Legendre polynomial.
    m : int
        The order of the associated Legendre polynomial.
    y : float, optional
        The value of sqrt(1 - x**2). If not provided, it is computed.

    Returns
    -------
    float
        The derivative of the associated Legendre polynomial of degree n and
        order m, evaluated at x.
    """
    y = y or sqrt_safe(1 - x**2)
    if m < 0:
        return (-1) ** (-m + 1) * u(x, n, -m)
    elif m == 0:
        return 0
    elif m == 1:
        return -1 / 4 * (3 / pi) ** 0.5
    elif n == m:
        return (
            -((n * (2 * n + 1) / (2 * (n + 1) * (n - 1))) ** 0.5)
            * y
            * u(x, n - 1, n - 1)
        )
    elif m == (n - 1):
        return ((2 * n + 1) * (n - 1) / (n + 1)) ** 0.5 * x * u(x, n - 1, n - 1)
    else:
        return ((2 * n + 1) * (n - 1) / ((n + 1) * (n**2 - m**2))) ** 0.5 * (
            x * (2 * n - 1) ** 0.5 * u(x, n - 1, m)
            - ((n - 2) * ((n - 1) ** 2 - m**2) / (n * (2 * n - 3))) ** 0.5
            * u(x, n - 2, m)
        )


def s(x, n, m, y=None):
    """
    Compute the derivative of the associated Legendre polynomial of degree n
    and order m, with respect to its argument x, evaluated at x.

    Parameters
    ----------
    x : float
        The argument of the associated Legendre polynomial.
    n : int
        The degree of the associated Legendre polynomial.
    m : int
        The order of the associated Legendre polynomial.
    y : float, optional
        The value of sqrt(1 - x**2). If not provided, it is computed.

    Returns
    -------
    float
        The derivative of the associated Legendre polynomial of degree n and
        order m, with respect to its argument x, evaluated at x.
    """

    y = y or sqrt_safe(1 - x**2)
    if m < 0:
        return (-1) ** (-m) * s(x, n, -m)
    elif n == m:
        return x * u(x, n, m)
    else:
        return 1 / (m + 1) * ((n + m + 1) * (n - m)) ** 0.5 * y * u(
            x, n, m + 1
        ) + x * u(x, n, m)


class SphericalSource(Source):
    """
    Spherical source.

    Parameters
    ----------
    wavelength : float
        The wavelength of the source.
    tol : float, optional
        The tolerance for the spherical Bessel function evaluation.
    **kwargs : dict
        Optional keyword arguments to be passed to the `Source` class.

    Attributes
    ----------
    tol : float
        The tolerance for the spherical Bessel function evaluation.
    r : dolfin.Expression
        The radial expression.
    cos_theta : dolfin.Expression
        The cosinus of the polar angle expression.
    sin_theta : dolfin.Expression
        The sinus of the polar angle expression.
    cos_phi : dolfin.Expression
        The cosinus of the azimuthal angle expression.
    sin_phi : dolfin.Expression
        The sinus of the azimuthal angle expression.
    """

    def __init__(self, wavelength, tol=1e-14, **kwargs):
        super().__init__(wavelength, dim=3, **kwargs)
        self.tol = tol
        r3D = "sqrt(pow(x[0],2) + pow(x[1],2) + pow(x[2],2))"
        r2D = "sqrt(pow(x[0],2) + pow(x[1],2))"
        self.r = df.Expression(
            f"{r3D}",
            **kwargs,
        )
        self.cos_theta = df.Expression(
            f"x[2]/({self.tol} + {r3D})",
            **kwargs,
        )
        self.sin_theta = df.Expression(
            f"{r2D}/({self.tol} + {r3D})",
            **kwargs,
        )
        self.cos_phi = df.Expression(
            f"x[0]/({self.tol} + {r2D})",
            **kwargs,
        )
        self.sin_phi = df.Expression(
            f"x[1]/({self.tol} + {r2D})",
            **kwargs,
        )
        ct = self.cos_theta
        st = self.sin_theta
        cp = self.cos_phi
        sp = self.sin_phi
        rot_mat = [
            [st * cp, ct * cp, -sp],
            [st * sp, ct * sp, cp],
            [ct, -st, df.Constant(0)],
        ]
        self.transfo_matrix = Complex(df.as_tensor(rot_mat))

    def phasor(self, m):
        if m == 0:
            return Constant(1 + 0j)
        else:
            return Complex(self.cos_phi, self.sin_phi) ** df.Constant(m)


class SphericalHarmonic(SphericalSource):
    """
    Spherical harmonic source class.

    This class defines a spherical harmonic source with a given degree `n` and order `m`.

    Parameters
    ----------
    wavelength : float
        The wavelength of the source.
    n : int
        The degree of the spherical harmonic.
    m : int
        The order of the spherical harmonic.
    **kwargs : dict
        Optional keyword arguments to be passed to the `Source` class.

    Attributes
    ----------
    n : int
        The degree of the spherical harmonic.
    m : int
        The order of the spherical harmonic.
    phm : dolfin.Expression
        The phasor of the spherical harmonic.
    components : dict
        The spherical components of the source.
    """

    def __init__(self, wavelength, n, m, **kwargs):
        super().__init__(wavelength, **kwargs)
        self.n = n
        self.m = m
        self.phm = self.phasor(self.m)
        self.components = {}

    @property
    def expression(self):
        return self.transfo_matrix * self.expression_spherical


class SphericalX(SphericalHarmonic):
    """
    SphericalX source class.

    Parameters
    ----------
    *args : tuple
        Positional arguments to be passed to the `SphericalHarmonic` class.
    **kwargs : dict
        Optional keyword arguments to be passed to the `SphericalHarmonic` class.

    Notes
    -----
    The spherical components of the source are defined as follows:

    - `r` component: 0
    - `theta` component: -j * u(cos(theta), n, m) * phm
    - `phi` component: s(cos(theta), n, m) * phm

    where `j` is the imaginary unit, `u` is the associated Legendre
    function, `s` is the derivative of the associated Legendre function with
    respect to the argument, and `phm` is the phasor of the spherical
    harmonic.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = Constant(0)
        self.components["theta"] = j * u(self.cos_theta, self.n, self.m) * self.phm
        self.components["phi"] = -s(self.cos_theta, self.n, self.m) * self.phm
        self.expression_spherical = vector(self.components.values())


class SphericalY(SphericalHarmonic):
    """
    SphericalY source class.

    Parameters
    ----------
    *args : tuple
        Positional arguments to be passed to the `SphericalHarmonic` class.
    **kwargs : dict
        Optional keyword arguments to be passed to the `SphericalHarmonic` class.

    Notes
    -----
    The spherical components of the source are defined as follows:

    - `r` component: P(cos(theta), n, m) * phm
    - `theta` component: 0
    - `phi` component: 0

    where `P` is the associated Legendre function and `phm` is the phasor of the spherical
    harmonic.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = P(self.cos_theta, self.n, self.m) * self.phm
        self.components["theta"] = Constant(0)
        self.components["phi"] = Constant(0)
        self.expression_spherical = vector(self.components.values())


class SphericalZ(SphericalHarmonic):
    """
    SphericalZ source class.

    Parameters
    ----------
    *args : tuple
        Positional arguments to be passed to the `SphericalHarmonic` class.
    **kwargs : dict
        Optional keyword arguments to be passed to the `SphericalHarmonic` class.

    Notes
    -----
    The spherical components of the source are defined as follows:

    - `r` component: 0
    - `theta` component: s(cos(theta), n, m) * phm
    - `phi` component: j * u(cos(theta), n, m) * phm

    where `j` is the imaginary unit, `u` is the associated Legendre
    function, `s` is the derivative of the associated Legendre function with
    respect to the argument, and `phm` is the phasor of the spherical
    harmonic.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = Constant(0)
        self.components["theta"] = s(self.cos_theta, self.n, self.m) * self.phm
        self.components["phi"] = j * u(self.cos_theta, self.n, self.m) * self.phm
        self.expression_spherical = vector(self.components.values())


class SphericalM(SphericalHarmonic):
    """
    SphericalM source class.

    Parameters
    ----------
    *args : tuple
        Positional arguments to be passed to the `SphericalHarmonic` class.
    **kwargs : dict
        Optional keyword arguments to be passed to the `SphericalHarmonic` class.

    Notes
    -----
    The spherical components of the source are defined as follows:

    - `r` component: 0
    - `theta` component: j * jn(n, kr) * Xnm
    - `phi` component: 0

    where `j` is the imaginary unit, `jn` is the spherical Bessel function of
    degree `n`, `kr` is the product of the wave number `k` and the radial
    distance `r`, and `Xnm` is the SphericalX source of degree `n` and order `m`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k = df.Constant(self.wavenumber)
        jn = sph_bessel_J(self.n, k * self.r)
        Xnm = SphericalX(*args, **kwargs)
        self.expression_spherical = jn * Xnm.expression_spherical


class SphericalN(SphericalHarmonic):
    """
    SphericalN source class.

    Parameters
    ----------
    *args : tuple
        Positional arguments to be passed to the `SphericalHarmonic` class.
    **kwargs : dict
        Optional keyword arguments to be passed to the `SphericalHarmonic` class.

    Notes
    -----
    The spherical components of the source are defined as a combination of
    the spherical Bessel function and the spherical harmonics:

    - `jn` is the spherical Bessel function of degree `n`.
    - `psin_prime` is the derivative of the Ricatti-Bessel function.
    - `Ynm` is the SphericalY source expression.
    - `Znm` is the SphericalZ source expression.

    These components are used to compute the expression for the SphericalN
    source.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k = df.Constant(self.wavenumber)
        jn = sph_bessel_J(self.n, k * self.r)
        psin_prime = ricatti_bessel_J(self.n - 1, k * self.r) - self.n * jn
        Y = SphericalY(*args, **kwargs)
        Z = SphericalZ(*args, **kwargs)
        Ynm = Y.expression_spherical
        Znm = Z.expression_spherical
        self.expression_spherical = (
            (self.n * (self.n + 1)) ** 0.5 * jn * Ynm + psin_prime * Znm
        ) / (k * self.r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import numpy as np
import pytest

from gyptis.utils.sample import _adaptive_sampler, adaptive_sampler

np.random.seed(13)


def test_sampler():
    Npoles = 100

    poles = np.random.rand(Npoles) + 1j * np.random.rand(Npoles)
    res = np.random.rand(Npoles) + 1j * np.random.rand(Npoles)

    def f(z):
        t = 0
        for p, r in zip(poles, res):
            t += r / (z - p)
        Q = np.abs(t) ** 2
        return Q / Npoles**2

    npts = 2000
    zref = np.linspace(0, 1, npts)
    f(zref)

    # adapt

    zmin, zmax = 0, 1
    n0 = 33
    z0 = np.linspace(zmin, zmax, n0)
    # t0 = f(z0)

    z, t = _adaptive_sampler(f, z0)
    f(z0)

    print(f"number of points: {len(z)}")
    print("------------")

    def f1(z):
        return f(z), z**2

    z, t = _adaptive_sampler(f1, z0)
    f(z0)
    assert len(z) == len(t)

    print(f"number of points: {len(z)}")

    @adaptive_sampler()
    def f2(z, a, b="test"):
        print(a)
        print(b)
        return f1(z)

    z1, t1 = f2(z0, 10, b="hello")
    assert len(z1) == len(t1)
    assert np.allclose(z1, z)
    assert np.allclose(t1, t)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .geometry import *


class Layered2D(Geometry):
    def __init__(
        self,
        period=1,
        thicknesses=OrderedDict(),
        pml_thickness=(1, 1),
        **kwargs,
    ):
        super().__init__(
            dim=2,
            **kwargs,
        )

        if isinstance(pml_thickness, (int, float)):
            pml_thickness = (pml_thickness, pml_thickness)
        thicknesses = thicknesses.copy()
        layer_names = list(thicknesses.keys())
        thicknesses["pml_y_" + layer_names[0]] = pml_thickness[0]
        thicknesses.move_to_end("pml_y_" + layer_names[0], last=False)

        thicknesses["pml_y_" + layer_names[-1]] = pml_thickness[1]
        self.pml_physical = ["pml_y_" + layer_names[0], "pml_y_" + layer_names[-1]]
        self.pml_thickness = pml_thickness
        self.period = period
        # assert isinstance(self.thicknesses == OrderedDict)
        self.thicknesses = thicknesses
        self.total_thickness = sum(self.thicknesses.values())

        self.y0 = -sum(list(self.thicknesses.values())[:2])
        self.layers = {}
        self.y_position = {}
        y0 = self.y0
        self._phys_groups = []
        for id, thickness in self.thicknesses.items():
            layer = self.make_layer(y0, thickness)
            self.layers[id] = layer
            self.y_position[id] = y0
            self.add_physical(layer, id)
            self._phys_groups.append(layer)
            # gmsh.model.removePhysicalName(id)
            y0 += thickness

        self.remove_all_duplicates()
        # print(self._phys_groups)
        for sub, num in self.subdomains_entities["surfaces"].items():
            self.add_physical(num, sub)

    @property
    def translation_x(self):
        return self._translation_matrix([self.period, 0, 0])

    def make_layer(self, y_position, thickness):
        return self.add_rectangle(
            -self.period / 2, y_position, 0, self.period, thickness
        )

    def build(self, *args, **kwargs):
        s = self.get_periodic_bnds(self.y0, self.total_thickness)
        periodic_id = {k: [S[-1] for S in v] for k, v in s.items()}
        gmsh.model.mesh.setPeriodic(
            1, periodic_id["+x"], periodic_id["-x"], self.translation_x
        )
        super().build(*args, **kwargs)

    def get_periodic_bnds(self, y_position, thickness, eps=1e-3):
        pmin = -self.period / 2 - eps, -eps + y_position, -eps
        pmax = -self.period / 2 + eps, y_position + thickness + eps, eps
        s = {"-x": gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 1)}
        pmin = +self.period / 2 - eps, -eps + y_position, -eps
        pmax = +self.period / 2 + eps, y_position + thickness + eps, eps
        s["+x"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 1)

        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from collections import OrderedDict
from pprint import pprint

import numpy as np

from gyptis import Grating, Layered, PlaneWave, dolfin


def test_grating3d(degree=1):
    dolfin.parameters["form_compiler"]["quadrature_degree"] = 2

    p = 1000

    #  ---------- incident wave ----------

    lambda0 = p * 3
    theta0 = 0 * np.pi / 180
    phi0 = 0 * np.pi / 180
    psi0 = 0 * np.pi / 180

    #  ---------- geometry ----------

    period = (p, p)
    grooove_thickness = p / 20
    hole_radius = p / 4
    thicknesses = OrderedDict(
        {
            "substrate": lambda0 / 1,
            "groove": grooove_thickness,
            "superstrate": lambda0 / 1,
        }
    )

    #  ---------- mesh ----------
    parmesh = 2

    parmesh_hole = parmesh * 1

    parmesh_groove = parmesh
    parmesh_pml = parmesh * 2 / 3

    mesh_params = dict(
        {
            "substrate": parmesh,
            "groove": parmesh_groove,
            "hole": parmesh_hole,
            "superstrate": parmesh,
        }
    )

    #  ---------- materials ----------
    eps_groove = (1.75 - 1.5j) ** 2
    eps_hole = 1
    eps_substrate = 1.5**2

    epsilon = dict(
        {
            "substrate": eps_substrate,
            "groove": eps_groove,
            "hole": eps_hole,
            "superstrate": 1,
        }
    )
    mu = dict({"substrate": 1, "groove": 1, "hole": 1, "superstrate": 1})

    index = {
        e[0]: np.mean(
            (np.array(e[1]).real.max() * np.array(m[1]).real.max()) ** 0.5
        ).real
        for e, m in zip(epsilon.items(), mu.items())
    }
    #  ---------- build geometry ----------
    geom = Layered(3, period, thicknesses, lambda0, finalize=False)

    groove = geom.layers["groove"]
    substrate = geom.layers["substrate"]
    superstrate = geom.layers["superstrate"]
    z0 = geom.z_position["groove"]

    hole = geom.add_cylinder(
        0,
        0,
        z0,
        0,
        0,
        z0 + grooove_thickness,
        hole_radius,
    )

    superstrate, substrate, hole, groove = geom.fragment(
        [superstrate, substrate, groove], hole
    )
    # hole, groove = geom.fragment(hole, groove)
    geom.add_physical(groove, "groove")
    geom.add_physical(hole, "hole")
    geom.add_physical(substrate, "substrate")
    geom.add_physical(superstrate, "superstrate")
    [geom.set_size(pml, lambda0 / parmesh_pml) for pml in geom.pml_physical]
    pmesh = {k: lambda0 / (index[k] * mesh_params[k]) for k in mesh_params}
    geom.set_mesh_size(pmesh)
    geom.build()

    pw = PlaneWave(
        wavelength=lambda0,
        angle=(theta0, phi0, psi0),
        dim=3,
        domain=geom.mesh,
        degree=degree,
    )
    bcs = {}
    s = Grating(
        geom,
        epsilon,
        mu,
        pw,
        boundary_conditions=bcs,
        degree=degree,
        periodic_map_tol=1e-6,
    )

    s.solve()
    print("  >> computing diffraction efficiencies")
    print("---------------------------------------")

    effs = s.diffraction_efficiencies(1, subdomain_absorption=True, orders=True)
    pprint(effs)
    R = np.sum(effs["R"])
    T = np.sum(effs["T"])
    Q = sum(q for t in effs["Q"].values() for q in t.values())
    print("sum R = ", R)
    print("sum T = ", T)
    print("sum Q = ", Q)
    print("B  = ", effs["B"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
Sources.
"""

from abc import ABC, abstractmethod

import numpy as np
import sympy as sp
from dolfin import Constant as ConstantRe
from scipy.constants import c, epsilon_0, mu_0
from sympy.vector import CoordSys3D

from .. import dolfin
from ..complex import Complex, Constant, as_tensor, as_vector, dot, grad
from ..plot import *

_COORD = CoordSys3D("N")


def sympyvector(components):
    """
    Convert a 3D vector to a sympy.vector expression.

    Parameters
    ----------
    components : array-like of shape (3,)
        The components of the vector.

    Returns
    -------
    vector : sympy.vector.Vector
        The sympy vector.
    """
    return (
        components[0] * _COORD.i + components[1] * _COORD.j + components[2] * _COORD.k
    )


x = sp.symbols("x[0] x[1] x[2]", real=True)
X = sympyvector(np.array(x))


def expression2complex_2d(expr, **kwargs):
    """
    Convert a 2D expression to a complex dolfin expression.

    Parameters
    ----------
    expr : sympy expression
        The 2D expression to convert.
    **kwargs : dict
        Keyword arguments to be passed to dolfin.Expression.

    Returns
    -------
    Complex
        The complex dolfin expression.
    """
    re, im = (p.subs(x[2], 0) for p in expr.as_real_imag())
    dexpr = [dolfin.Expression(sp.printing.ccode(p), **kwargs) for p in (re, im)]
    return Complex(*dexpr)


class Source(ABC):
    """
    Abstract base class for defining a source in the gyptis framework.

    Parameters
    ----------
    wavelength : float
        The wavelength of the source.
    dim : int
        The dimension of the source.
    phase : float, optional
        The phase shift of the source. Default is 0.
    amplitude : float, optional
        The amplitude of the source. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    """

    def __init__(self, wavelength, dim, phase=0, amplitude=1, degree=1, domain=None):
        self.wavelength = wavelength
        self.phase = phase
        self.amplitude = amplitude
        self.dim = dim
        self.degree = degree
        self.domain = domain

    @property
    def wavenumber(self):
        """
        Calculate the wave number from the wavelength.

        Returns
        -------
        float
            The wave number calculated using the formula 2 * pi / wavelength.
        """

        return 2 * np.pi / self.wavelength

    @property
    def pulsation(self):
        """
        The pulsation of the source, calculated from the wave number and speed of light.

        Returns
        -------
        float
            The pulsation of the source.
        """
        return self.wavenumber * c

    @property
    def frequency(self):
        """
        The frequency of the source, calculated from the wavelength.

        Returns
        -------
        float
            The frequency of the source.
        """
        return c / self.wavelength

    @abstractmethod
    def expression(self):
        """
        Compute the expression for the source.

        This method must be implemented by subclasses to provide the specific
        expression of the source in the framework.

        Returns
        -------
        dolfin.Expression
            The Dolfin expression representing the source.
        """

        pass

    def plot(self, figsize=None, ax=None):
        """
        Plot the source expression.

        Parameters
        ----------
        figsize : tuple of int, optional
            The size of the figure in inches. If None, the size of the figure is
            not modified.
        ax : matplotlib.axes.Axes, optional
            The axes to plot on. If None, a new figure and axes are created.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure containing the plot.
        ax : matplotlib.axes.Axes
            The axes containing the plot.
        tri : matplotlib.tri.Triangulation
            The triangulation of the mesh.
        cb : matplotlib.colorbar.Colorbar
            The colorbar of the plot.
        """
        if self.dim == 2:
            tri, cb = plot(self.expression, mesh=self.domain, ax=ax)
            fig = plt.gcf()
            ax = fig.axes
            ax_ = ax[:2]
            ax_[0].set_title("Re")
            ax_[1].set_title("Im")
            for a in ax_:
                a.set_xlabel("x")
                a.set_ylabel("y")
            if figsize is not None:
                fig.set_size_inches(figsize)
            plt.tight_layout()
            return fig, ax, tri, cb
        else:
            raise NotImplementedError("plot not implemented in 3D")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from gyptis.geometry import *


class LayeredBoxPML3D(Geometry):
    def __init__(
        self,
        width,
        thicknesses,
        pml_width=None,
        **kwargs,
    ):
        super().__init__(
            dim=3,
            **kwargs,
        )

        # self.cross_section = LayeredBoxPML2D(
        #     width[0], thicknesses, (pml_width[0], pml_width[1]), **kwargs
        # )

        self.build_pmls = pml_width != None
        self.build_pmls_z = pml_width[2] != None

        self.width = width
        self.thicknesses = thicknesses
        self.total_thickness = sum(self.thicknesses.values())

        self.layers = {}
        self.y_position = {}
        self.box_size = width[0], self.total_thickness, width[1]
        self.pml_width = pml_width

        y0 = -self.total_thickness / 2
        if self.build_pmls:
            tx0 = self.pml_width[0] / 2 + self.box_size[0] / 2
            if self.build_pmls_z:
                tz0 = self.pml_width[2] / 2 + self.box_size[2] / 2

        for name, thickness in self.thicknesses.items():
            layer = self.make_layer(y0, thickness)
            self.layers[name] = layer
            self.y_position[name] = y0
            self.add_physical(layer, name)
            if self.build_pmls:
                spml = (self.pml_width[0], thickness, self.width[1])
                ty = thickness / 2 + y0
                _pmls = []
                for i in [-1, 1]:
                    t = np.array([i * tx0, ty, 0])
                    _pml = self._add_box_translate(spml, t)
                    _pmls.append(_pml)
                self.add_physical(_pmls, "pml_x_" + name)
                if self.build_pmls_z:
                    spml = (self.width[0], thickness, self.pml_width[2])
                    _pmls = []
                    for i in [-1, 1]:
                        t = np.array([0, ty, i * tz0])
                        _pml = self._add_box_translate(spml, t)
                        _pmls.append(_pml)
                    self.add_physical(_pmls, "pml_z_" + name)

                    spml = (self.pml_width[0], thickness, self.pml_width[2])
                    _pmls = []
                    for i in [-1, 1]:
                        for j in [-1, 1]:
                            t = np.array([i * tx0, ty, j * tz0])
                            _pml = self._add_box_translate(spml, t)
                            _pmls.append(_pml)
                    self.add_physical(_pmls, "pml_xz_" + name)

            y0 += thickness

        if self.build_pmls:
            y0 = -self.total_thickness / 2
            names = list(self.layers.keys())

            top_bot_pos = [
                -self.pml_width[1] / 2 + y0,
                self.total_thickness + y0 + self.pml_width[1] / 2,
            ]

            for j in [0, -1]:
                tz = np.array([0, top_bot_pos[j], 0])
                spml = (self.box_size[0], self.pml_width[1], self.box_size[2])
                _pml = self._add_box_translate(spml, tz)
                self.add_physical(_pml, "pml_y_" + names[j])

                spml = (self.pml_width[0], self.pml_width[1], self.box_size[2])
                _pmls = []
                for i in [-1, 1]:
                    t = np.array([i * tx0, top_bot_pos[j], 0])
                    _pml = self._add_box_translate(spml, t)
                    _pmls.append(_pml)
                self.add_physical(_pmls, "pml_xy_" + names[j])

                if self.build_pmls_z:
                    spml = (self.box_size[0], self.pml_width[1], self.pml_width[2])
                    _pmls = []
                    for i in [-1, 1]:
                        t = np.array([0, top_bot_pos[j], i * tz0])
                        _pml = self._add_box_translate(spml, t)
                        _pmls.append(_pml)
                    self.add_physical(_pmls, "pml_yz_" + names[j])

                    spml = (self.pml_width[0], self.pml_width[1], self.pml_width[2])
                    _pmls = []
                    for i in [-1, 1]:
                        for k in [-1, 1]:
                            t = np.array([i * tx0, top_bot_pos[j], k * tz0])
                            _pml = self._add_box_translate(spml, t)
                            _pmls.append(_pml)
                    self.add_physical(_pmls, "pml_xyz_" + names[j])

        self.remove_all_duplicates()

        for sub, num in self.subdomains_entities["volumes"].items():
            self.add_physical(num, sub)

        self._initial_setup = self._get_initial_setup()
        self._subdomains_ids = self._initial_setup[1]

    @property
    def subdomains_ids(self):
        return self._subdomains_ids

    @subdomains_ids.setter
    def subdomains_ids(self, subdomain_ids):
        self._restore_initial_subdomains(subdomain_ids)

    def _add_box_center(self, box_size):
        corner = -np.array(box_size) / 2
        return self.add_box(*corner, *box_size)

    def _translate(self, tag, t):
        translation = tuple(t)
        self.translate(self.dimtag(tag), *translation)

    def _add_box_translate(self, box_size, translation):
        pml = self._add_box_center(box_size)
        self._translate(pml, translation)
        return pml

    def make_layer(self, y_position, thickness):
        return self.add_box(
            -self.width[0] / 2,
            y_position,
            -self.width[1] / 2,
            self.width[0],
            thickness,
            self.width[1],
        )

    def _get_initial_setup(self):
        all_subdomains = list(self.subdomains_entities["volumes"].values())
        all_subdomains_flat = [item for sublist in all_subdomains for item in sublist]
        lengths = [len(sublist) for sublist in all_subdomains]
        all_subdomains_name = list(self.subdomains_entities["volumes"].keys())
        return all_subdomains_name, all_subdomains_flat, lengths

    def _restore_initial_subdomains(self, subdomain_ids):
        all_subdomains_name, _, lengths = self._initial_setup
        i = 0
        all_subdomains_restored = []
        for n in lengths:
            all_subdomains_restored.append(subdomain_ids[i : i + n])
            i += n
        for name, num in zip(all_subdomains_name, all_subdomains_restored):
            self.add_physical(num, name)


from collections import OrderedDict

import matplotlib.pyplot as plt

import gyptis as gy

plt.close("all")
plt.ion()

box_width = 8, 6
pml_width = 1, 1.2, 2
pml_width = 1, 1.2, None


wl = 1

nsub = 1.43
nsup = 1
ncore = 4
wg_width = 1
wg_thickness = 0.1
wg_depth = 2

hsup = hsub = wg_thickness * 10

pmesh = 2
lmin = wl / pmesh

thicknesses = OrderedDict(substrate=hsub, superstrate=hsup)

geom3D = LayeredBoxPML3D(box_width, thicknesses=thicknesses, pml_width=pml_width)


mid_width = wg_width * 2
mid_thickness = wg_thickness * 3
mid_depth = wg_depth / 4

ysub = geom3D.y_position["superstrate"]

design_size = 4

wg_width = 1
wg_depth1 = geom3D.box_size[2] / 2 - design_size / 2
wg_depth2 = geom3D.box_size[0] / 2 - design_size / 2

# core1 = geom3D.add_box(
#     -wg_width / 2,
#     ysub,
#     -geom3D.box_size[2]/2- geom3D.pml_width[2],
#     wg_width,
#     wg_thickness,
#     wg_depth1 + geom3D.pml_width[2],
# )

# # core2 = geom3D.add_box(
# #     design_size/2,
# #     ysub,
# #     -wg_width / 2,
# #     wg_depth2+ geom3D.pml_width[0],
# #     wg_thickness,
# #     wg_width,
# # )
# core2 = geom3D.add_box(
#     -wg_width / 2,
#     ysub,
#     design_size/2,
#     wg_width,
#     wg_thickness,
#     wg_depth1+ geom3D.pml_width[2],
# )


core1 = geom3D.add_box(
    -wg_width / 2,
    ysub,
    -geom3D.box_size[2] / 2,
    wg_width,
    wg_thickness,
    wg_depth1,
)

core2 = geom3D.add_box(
    -wg_width / 2,
    ysub,
    design_size / 2,
    wg_width,
    wg_thickness,
    wg_depth1,
)


design = geom3D.add_box(
    -design_size / 2,
    ysub,
    -design_size / 2,
    design_size,
    wg_thickness,
    design_size,
)

# fragments = geom3D.fragment([core1,design,core2], geom3D.subdomains_ids)
# core1 = fragments[1]
# design = fragments[2]
# core2 = fragments[3]
# geom3D.subdomains_ids = fragments[5:]

# geom3D.add_physical([core1,core2], "core")
# geom3D.add_physical([design], "design")
# # geom3D.add_physical([fragments[0]], "pml_z_core")
# # geom3D.add_physical([fragments[4]], "pml_x_core")
# geom3D.add_physical([fragments[0], fragments[4]], "pml_z_core")


fragments = geom3D.fragment([core1, design, core2], geom3D.subdomains_ids)
core1 = fragments[0]
design = fragments[1]
core2 = fragments[2]
geom3D.subdomains_ids = fragments[3:]

geom3D.add_physical([core1, core2], "core")
geom3D.add_physical([design], "design")

print("Building 3D mesh")
geom3D.build(1)
print("Done 3D mesh")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


"""
This module provides various source implementations for electromagnetic simulations in the gyptis library.

The sources include different types of electromagnetic waveforms such as dipole, Gaussian beam, line source,
plane wave, and spherical harmonics. These sources are essential for defining the incident fields in
computational simulations using the gyptis framework.

Modules:
- dipole: Defines a Dipole source for 2D electromagnetic simulations.
- gaussian: Implements a Gaussian beam source.
- ls: Contains the LineSource class for representing line sources.
- pw: Provides the PlaneWave class for plane wave sources.
- source: Abstract base class for all sources, defining common properties and methods.
- spharm: Implements spherical harmonics and related operations.
- stack: Utilities for stacking and manipulating sources.

Each module is designed to be used in the context of finite element method (FEM) simulations, with support
for 2D and partially for 3D geometries.
"""

from .dipole import *
from .gaussian import *
from .ls import *
from .pw import *
from .source import *
from .spharm import *
from .stack import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


__all__ = [
    "array2function",
    "function2array",
    "project_iterative",
    "get_coordinates",
    "rot_matrix_2d",
    "tanh",
]


import numpy as np

from .. import ADJOINT, dolfin
from ..complex import project


def array2function(values, function_space):
    """Convert a numpy array to a fenics Function.

    Parameters
    ----------
    values : numpy array
        The array to convert.
    function_space : FunctionSpace
        The function space to interpolate on.

    Returns
    -------
    Function
        The converted array.

    """

    # print(len(values))

    if np.isscalar(values):
        values = np.ones(function_space.dim()) * values

    u = dolfin.Function(function_space)
    dofmap = function_space.dofmap()
    # order = dofmap.tabulate_local_to_global_dofs()

    dofmap = function_space.dofmap()
    my_first, my_last = dofmap.ownership_range()
    sub_array = values[my_first:my_last]  # [order]
    # print("order", len(order))
    u.vector().set_local(sub_array)
    dolfin.as_backend_type(u.vector()).update_ghost_values()
    # dolfin.as_backend_type(u.vector()).vec().ghostUpdate()
    # u.vector().apply("")
    u.vector().apply("insert")
    return u


#
# def array2function(values, function_space):
#     u = dolfin.Function(function_space)
#     u.set_allow_extrapolation(True)
#     u.vector().apply("insert")
#     u.vector().set_local(values)
#     dolfin.as_backend_type(u.vector()).update_ghost_values()
#     u.vector().apply("insert")
#     return u
#
#
#
# def array2function(values, function_space):
#     V = function_space
#     u = dolfin.Function(function_space)
#     vec = u.vector()
#     mesh = function_space.mesh
#
#     dofmap = V.dofmap()
#     my_first, my_last = dofmap.ownership_range()                # global
#
#     # # 'Handle' API change of tabulate coordinates
#     # if df.dolfin_version().split('.')[1] == '7':
#     #     x = V.tabulate_dof_coordinates().reshape((-1, 2))
#     # else:
#     # x = V.dofmap().tabulate_all_coordinates(mesh)
#     #
#     x = V.tabulate_dof_coordinates().reshape((-1, 2))
#
#     unowned = dofmap.local_to_global_unowned()
#     dofs = [dofmap.local_to_global_index(dof) not in unowned for dof in range(my_last-my_first)]
#     x = x[dofs]
#     vec.set_local(values)
#     vec.apply('insert')
#     return u


def function2array(f, space=None):
    """Convert a fenics Function to a numpy array.

    Parameters
    ----------
    f : Function
        The function to convert.

    Returns
    -------
    numpy array
        The converted function.

    """
    if space is not None:
        f = project_iterative(f, space)
    values = f.vector().get_local()
    function_space = f.function_space()
    dofmap = function_space.dofmap()
    my_first, my_last = dofmap.ownership_range()
    return values


def project_iterative(applied_function, function_space):
    kwargs = dict(solver_type="cg", preconditioner_type="jacobi")
    if ADJOINT:
        kwargs["annotate"] = False
    return project(applied_function, function_space, **kwargs)


def get_coordinates(A):
    n = A.dim()
    d = A.mesh().geometry().dim()
    dof_coordinates = A.tabulate_dof_coordinates()
    dof_coordinates.resize((n, d))
    return dof_coordinates


def rot_matrix_2d(t):
    return np.array([[np.sin(t), -np.cos(t), 0], [np.cos(t), np.sin(t), 0], [0, 0, 1]])


def tanh(x):
    return (dolfin.exp(2 * x) - 1) / (dolfin.exp(2 * x) + 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import os
import tempfile

import meshio
import numpy as np
from dolfin.cpp.mesh import MeshFunctionSizet

from .. import ADJOINT, dolfin


def read_mesh(mesh_file, data_dir=None, data_dir_xdmf=None, dim=3, subdomains=None):
    data_dir_xdmf = data_dir_xdmf or tempfile.mkdtemp()
    meshio_mesh = meshio.read(mesh_file, file_format="gmsh")
    if dim == 3:
        base_cell_type = "tetra"
    elif dim == 2:
        base_cell_type = "triangle"
    else:
        base_cell_type = "line"

    points = meshio_mesh.points if dim == 3 else meshio_mesh.points[:, :2]
    physicals = meshio_mesh.cell_data_dict["gmsh:physical"]

    cell_types, data_gmsh = zip(*physicals.items())
    data_gmsh = list(data_gmsh)
    cells = {ct: [] for ct in cell_types}

    for cell_type in cell_types:
        for cell in meshio_mesh.cells:
            if cell.type == cell_type:
                cells[cell_type].append(cell.data)
        cells[cell_type] = np.vstack(cells[cell_type])

    icell = np.where(np.array(cell_types) == base_cell_type)[0][0]
    if subdomains is not None:
        doms = subdomains if hasattr(subdomains, "__len__") else [subdomains]
        mask = np.hstack([np.where(data_gmsh[icell] == i) for i in doms])[0]
        data_gmsh_ = data_gmsh[icell][mask]
        data_gmsh[icell] = data_gmsh_
        cells[base_cell_type] = cells[base_cell_type][mask]

    mesh_data = {}

    for cell_type, data in zip(cell_types, data_gmsh):
        meshio_data = meshio.Mesh(
            points=points,
            cells={cell_type: cells[cell_type]},
            cell_data={cell_type: [data]},
        )
        meshio.xdmf.write(f"{data_dir_xdmf}/{cell_type}.xdmf", meshio_data)
        mesh_data[cell_type] = meshio_data

    dolfin_mesh = dolfin.Mesh()
    with dolfin.XDMFFile(
        dolfin_mesh.mpi_comm(), f"{data_dir_xdmf}/{base_cell_type}.xdmf"
    ) as infile:
        infile.read(dolfin_mesh)

    markers = {}
    dim_map = dict(vertex=0, line=1, triangle=2, tetra=3)
    if subdomains is not None:
        cell_types = [base_cell_type]
    for cell_type in cell_types:
        mvc = dolfin.MeshValueCollection("size_t", dolfin_mesh, dim_map[cell_type])
        try:
            with dolfin.XDMFFile(
                dolfin_mesh.mpi_comm(), f"{data_dir_xdmf}/{cell_type}.xdmf"
            ) as infile:
                infile.read(mvc, cell_type)
        except Exception:
            pass

        markers[cell_type] = MeshFunctionSizet(dolfin_mesh, mvc)

    return dict(mesh=dolfin_mesh, markers=markers)


class MarkedMesh:
    def __init__(self, filename, geometric_dimension=3, data_dir=None):
        self.data_dir = data_dir
        self.geometric_dimension = geometric_dimension
        self.filename = filename
        data_dir = data_dir or tempfile.mkdtemp()
        dic = read_mesh(filename, dim=geometric_dimension, data_dir=data_dir)
        self.mesh = dic["mesh"]
        self.markers = dic["markers"]
        self.dimension = self.mesh.geometric_dimension()


def run_submesh(geom, subdomain, outpath=None):
    outpath = outpath or os.path.join(tempfile.mkdtemp(), "submesh.xdmf")
    fname = os.path.join(os.path.dirname(__file__), "_submesh.py")
    meshname = os.path.join(geom.data_dir, geom.mesh_name)
    domnum = geom.domains[subdomain]
    dim = geom.dim
    os.system(f"mpirun -n 1 python {fname} {meshname} {domnum} {dim} {outpath}")
    return outpath


def read_xdmf_mesh(outpath):
    dolfin_mesh = dolfin.Mesh()
    with dolfin.XDMFFile(dolfin_mesh.mpi_comm(), outpath) as infile:
        infile.read(dolfin_mesh)
    return dolfin_mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import pytest

from gyptis.models import *

a = 1
v = (a, 0), (0, a)
R1 = 0.3 * a
R2 = 0.4 * a
lmin = 0.1

lattice = Lattice2D(v)
circ = lattice.add_ellipse(a / 2, a / 2, 0, R1, R2)
circ, cell = lattice.fragment(circ, lattice.cell)
lattice.add_physical(cell, "background")
lattice.add_physical(circ, "inclusion")
lattice.set_size("background", lmin)
lattice.set_size("inclusion", lmin)
lattice.build()


@pytest.mark.parametrize(
    "degree,epsincl,muincl",
    [(1, 4 - 3j, 1), (2, 4 - 3j, 1), (1, 4 - 3j, 3 - 0.1j), (2, 4 - 3j, 3 - 0.1j)],
)
def test_hom(degree, epsincl, muincl):
    EPS = []
    MU = []
    for eps_inclusion, mu_inclusion in zip(
        [epsincl, epsincl * np.eye(3)], [muincl, muincl * np.eye(3)]
    ):
        epsilon = dict(background=1.25, inclusion=eps_inclusion)
        mu = dict(background=1, inclusion=mu_inclusion)

        hom = Homogenization2D(
            lattice,
            epsilon,
            mu,
            degree=degree,
        )
        eps_eff = hom.get_effective_permittivity()
        EPS.append(eps_eff)
        mu_eff = hom.get_effective_permeability()
        MU.append(mu_eff)

    assert np.allclose(*EPS)
    assert np.allclose(*MU)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from .source import *


def green_function_2d(wavelength, xs, ys, phase=0, amplitude=1, degree=1, domain=None):
    """
    Compute the 2D Green function associated with a point source.

    Parameters
    ----------
    wavelength : float
        The wavelength of the Green function.
    xs : float
        The x-coordinate of the point source.
    ys : float
        The y-coordinate of the point source.
    phase : float, optional
        The phase shift of the Green function. Default is 0.
    amplitude : float, optional
        The amplitude of the Green function. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The Green function as a dolfin Expression.
    """

    Xs = sympyvector(sp.symbols("xs, ys, 0", real=True))
    k0 = sp.symbols("k0", real=True)
    Xshift = X - Xs
    rho = sp.sqrt(Xshift.dot(Xshift))
    rho = rho.subs(x[2], 0)
    kr = k0 * rho
    k0_ = 2 * np.pi / wavelength
    KR = dolfin.Expression(
        sp.printing.ccode(kr), k0=k0_, xs=xs, ys=ys, degree=degree, domain=domain
    )
    re = dolfin.Expression("y0(KR)", KR=KR, degree=degree, domain=domain)
    im = dolfin.Expression("j0(KR)", KR=KR, degree=degree, domain=domain)
    return (
        -1
        / 4
        * Complex(re, im)
        * Constant(amplitude)
        * phase_shift_constant(ConstantRe(phase))
    )


class LineSource(Source):
    """
    LineSource class.

    Parameters
    ----------
    wavelength : float
        The wavelength of the line source.
    position : tuple of float
        The (x, y) coordinates of the line source position.
    dim : int, optional
        The dimension of the line source. Default is 2.
    phase : float, optional
        The phase shift of the line source. Default is 0.
    amplitude : float, optional
        The amplitude of the line source. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Raises
    ------
    NotImplementedError
        If the dimension `dim` is 3, since LineSource is not implemented in 3D.
    """

    def __init__(
        self, wavelength, position, dim=2, phase=0, amplitude=1, degree=1, domain=None
    ):
        if dim == 3:
            raise NotImplementedError("LineSource not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.position = position

    @property
    def expression(self):
        """
        The expression of the line source.

        Returns
        -------
        expr : Expression
            The expression of the line source as a dolfin Expression.
        """
        return green_function_2d(
            self.wavelength,
            *self.position,
            phase=self.phase,
            amplitude=self.amplitude,
            degree=self.degree,
            domain=self.domain,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import copy

from .geometry import *


class LayeredBoxPML2D(Geometry):
    def __init__(
        self,
        width,
        thicknesses,
        pml_width,
        **kwargs,
    ):
        super().__init__(
            dim=2,
            **kwargs,
        )
        self._init_args = {
            "width": width,
            "thicknesses": thicknesses,
            "pml_width": pml_width,
            **kwargs,
        }

        self.width = width
        self.thicknesses = thicknesses
        self.total_thickness = sum(self.thicknesses.values())

        self.layers = {}
        self.y_position = {}
        self.box_size = width, self.total_thickness
        self.pml_width = pml_width

        def _addrect_center(rect_size):
            corner = -np.array(rect_size) / 2
            corner = tuple(corner) + (0,)
            return self.add_rectangle(*corner, *rect_size)

        def _translate(tag, t):
            translation = tuple(t) + (0,)
            self.translate(self.dimtag(tag), *translation)

        def _add_pml(rect_size, translation):
            pml = _addrect_center(rect_size)
            _translate(pml, translation)
            return pml

        y0 = -self.total_thickness / 2
        tx0 = self.pml_width[0] / 2 + self.box_size[0] / 2

        for name, thickness in self.thicknesses.items():
            layer = self.make_layer(y0, thickness)
            self.layers[name] = layer
            self.y_position[name] = y0
            self.add_physical(layer, name)
            sx = (self.pml_width[0], thickness)
            ty0 = thickness / 2 + y0
            tx = np.array([tx0, ty0])
            pmlxp = _add_pml(sx, tx)

            tx = np.array([-tx0, ty0])
            pmlxm = _add_pml(sx, tx)
            self.add_physical([pmlxp, pmlxm], "pml_x_" + name)

            y0 += thickness

        sy = (self.box_size[0], self.pml_width[1])
        ty = np.array([0, self.pml_width[1] / 2 + self.box_size[1] / 2])

        pmlyp = _add_pml(sy, ty)
        pmlym = _add_pml(sy, -ty)

        names = list(self.layers.keys())
        self.add_physical(pmlym, "pml_y_" + names[0])
        self.add_physical(pmlyp, "pml_y_" + names[-1])

        s = (self.pml_width[0], self.pml_width[1])
        t = np.array(
            [
                self.pml_width[0] / 2 + self.box_size[0] / 2,
                self.pml_width[1] / 2 + self.box_size[1] / 2,
            ]
        )
        pmlxypp = _add_pml(s, t)
        pmlxymm = _add_pml(s, -t)
        pmlxymp = _add_pml(s, (-t[0], t[1]))
        pmlxypm = _add_pml(s, (t[0], -t[1]))

        self.add_physical([pmlxypm, pmlxymm], "pml_xy_" + names[0])
        self.add_physical([pmlxypp, pmlxymp], "pml_xy_" + names[-1])

        self.pmls = []
        for name in names:
            self.pmls.append("pml_x_" + name)
        for name in [names[-1], names[0]]:
            self.pmls.append("pml_y_" + name)
            self.pmls.append("pml_xy_" + name)

        self.remove_all_duplicates()

        for sub, num in self.subdomains_entities["surfaces"].items():
            self.add_physical(num, sub)

        x0 = self.width / 2 + self.pml_width[0]
        y0 = (
            list(self.y_position.values())[-1]
            + list(self.thicknesses.values())[-1]
            + self.pml_width[1]
        )
        y1 = list(self.y_position.values())[0] - self.pml_width[1]

        horizontal_lines = self.find_lines_near_positions("x", [-x0, x0])
        vertical_lines = self.find_lines_near_positions("y", [y0, y1])
        outer_bnds = horizontal_lines + vertical_lines
        self.add_physical(outer_bnds, "outer_boundaries", dim=1)
        self._initial_subdomains_entities = copy.deepcopy(self.subdomains_entities)
        self._initial_subdomains = copy.deepcopy(self.subdomains)

    @property
    def new_subdomains(self):
        initial = self._initial_subdomains
        current = self.subdomains
        diff = {}

        for section in initial:  # 'volumes', 'surfaces', ...
            initial_sub = initial[section]
            current_sub = current[section]

            new_items = {k: v for k, v in current_sub.items() if k not in initial_sub}

            # always keep the section, even if empty
            diff[section] = new_items

        return diff

    def make_layer(self, y_position, thickness):
        return self.add_rectangle(-self.width / 2, y_position, 0, self.width, thickness)

    def find_lines_near_positions(self, axis, positions, tol_dir=1e-12, tol_pos=1e-12):
        """
        Generic helper for finding axis-aligned lines near given coordinates.

        axis = "x"  -> find horizontal lines near x = x0
        axis = "y"  -> find vertical   lines near y = y0
        """
        assert axis in ("x", "y"), "axis must be 'x' or 'y'"
        results = []

        for dim, tag in self.get_entities(dim=1):
            nodes = self.get_boundaries(tag, dim=1, physical=False)
            pts = [self.model.get_value(0, n, []) for n in nodes]
            if len(pts) != 2:
                continue
            (x1, y1, _), (x2, y2, _) = pts

            if axis == "y":
                # horizontal lines near x = x0
                for y0 in positions:
                    if y1 == y2 == y0:
                        results.append(tag)

            else:
                # horizontal lines near x = x0
                for x0 in positions:
                    if x1 == x2 == x0:
                        results.append(tag)
        return results

    def copy_instance(self):
        return self.__class__(**self._init_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from ..sources import spharm as sh
from .metaclasses import _ScatteringBase
from .simulation import *


class Scatt3D(_ScatteringBase, Simulation):
    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        source=None,
        boundary_conditions=None,
        polarization=None,
        modal=False,
        degree=1,
        pml_stretch=1 - 1j,
        element="N1curl",
        **kwargs,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, BoxPML3D)
        assert source.dim == 3
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)
        function_space = ComplexFunctionSpace(geometry.mesh, element, degree)

        names = geometry.domains.keys()
        pmls_list = init_pmls(names, pml_stretch)

        epsilon_coeff = Coefficient(
            self.epsilon,
            geometry,
            pmls=pmls_list,
            degree=degree,
            dim=3,
        )
        mu_coeff = Coefficient(
            self.mu,
            geometry,
            pmls=pmls_list,
            degree=degree,
            dim=3,
        )

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["box"] + geometry.pml_physical
        source_domains = [
            dom for dom in geometry.domains if dom not in no_source_domains
        ]

        formulation = Maxwell3D(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="box",
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation, **kwargs)

        self.Z0 = np.sqrt(mu_0 / epsilon_0)
        self.S0 = 1 / (2 * self.Z0)
        self.degree = degree

    def solve_system(self, again=False):
        E = super().solve_system(again=again, vector_function=False)
        self.solution = {"diffracted": E, "total": E + self.source.expression}
        return E

    def _cross_section_helper(self, return_type="s", boundaries="calc_bnds"):
        parallel = dolfin.MPI.comm_world.size > 1
        # normal vector is messing up in parallel so workaround here:
        if parallel:
            Rcalc = self.geometry.Rcalc
            n_out = dolfin.Expression(
                (f"-x[0]/{Rcalc}", f"-x[1]/{Rcalc}", f"-x[2]/{Rcalc}"),
                degree=self.degree,
            )
        else:
            n_out = self.geometry.unit_normal_vector

        Es = self.solution["diffracted"]
        omega = self.source.pulsation
        inv_mu_coeff = self.formulation.mu.invert().as_subdomain()
        Hs = inv_mu_coeff / Complex(0, dolfin.Constant(omega * mu_0)) * curl(Es)
        Ss = dolfin.Constant(0.5) * cross(Es, Hs.conj).real

        Ei = self.source.expression
        mu_a = self.formulation.mu.build_annex(
            domains=self.formulation.source_domains,
            reference=self.formulation.reference,
        )
        inv_mua_coeff = mu_a.invert().as_subdomain()
        Hi = inv_mua_coeff / Complex(0, dolfin.Constant(omega * mu_0)) * curl(Ei)
        # Si = dolfin.Constant(0.5) * cross(Ei, Hi.conj).real
        Se = dolfin.Constant(0.5) * (cross(Ei, Hs.conj) + cross(Es, Hi.conj)).real

        Etot = self.solution["total"]
        Htot = inv_mu_coeff / Complex(0, dolfin.Constant(omega * mu_0)) * curl(Etot)
        Stot = dolfin.Constant(0.5) * cross(Etot, Htot.conj).real

        if return_type == "s":
            Ws = assemble(dot(n_out("+"), Ss("+")) * self.dS(boundaries))
            return Ws / self.S0
        if return_type == "e":
            We = -assemble(dot(n_out("+"), Se("+")) * self.dS(boundaries))
            return We / self.S0
        if return_type == "a":
            Wa = -assemble(dot(n_out("+"), Stot("+")) * self.dS(boundaries))
            return Wa / self.S0

    def scattering_cross_section(self, **kwargs):
        return self._cross_section_helper("s", **kwargs)

    def extinction_cross_section(self, **kwargs):
        return self._cross_section_helper("e", **kwargs)

    def absorption_cross_section(self, **kwargs):
        return self._cross_section_helper("a", **kwargs)

    def get_T_matrix_coeff(
        self,
        p0,
        p1,
        source_type,
        coeff_type,
        solve_again=False,
        degree_source=None,
        component="transverse",
        boundary="calc_bnds",
    ):
        if coeff_type not in ["E", "H", "both"]:
            raise ValueError("coeff_type must be E, H or both")
        if component not in ["normal", "transverse"]:
            raise ValueError("component must be either normal or transverse")
        degree_source = degree_source or self.degree
        Rcalc = self.geometry.Rcalc
        wavelength = self.source.wavelength
        n0, m0 = sh.p2nm(p0)
        n1, m1 = sh.p2nm(p1)
        mesh = self.geometry.mesh
        if source_type == "N":
            source = sh.SphericalN(
                wavelength,
                n0,
                m0,
                domain=mesh,
                degree=degree_source,
            )
        else:
            source = sh.SphericalM(
                wavelength,
                n0,
                m0,
                domain=mesh,
                degree=degree_source,
            )

        self.source = source

        if not solve_again:
            self.solve()
        else:
            self.assemble_rhs()
            self.solve_system(again=solve_again)

        Escat = self.solution["diffracted"]
        k = source.wavenumber

        def _postpro_coeff_e():
            if component == "normal":
                # normal components
                Y = sh.SphericalY(
                    wavelength,
                    n1,
                    m1,
                    domain=mesh,
                    degree=degree_source,
                )
                integrand = dot(Escat, Y.expression.conj)
                integral = assemble(
                    0.5 * (integrand("+")) * self.dS(boundary) + Constant(0) * self.dx
                ) / (Rcalc**2)
                coeff = (
                    integral
                    * k
                    * Rcalc
                    / (sh.sph_hn2(n1, k * Rcalc) * (n1 * (n1 + 1)) ** 0.5)
                )
            else:
                # transverse components
                Z = sh.SphericalZ(
                    wavelength,
                    n1,
                    m1,
                    domain=mesh,
                    degree=degree_source,
                )
                integrand = dot(Escat, Z.expression.conj)
                integral = assemble(
                    0.5 * (integrand("+")) * self.dS(boundary) + Constant(0) * self.dx
                ) / (Rcalc**2)
                xsi_prime = sh.rb_hn2(n1 - 1, k * Rcalc) - n1 * sh.sph_hn2(
                    n1, k * Rcalc
                )
                coeff = integral * k * Rcalc / xsi_prime
            return coeff

        def _postpro_coeff_h():
            X = sh.SphericalX(
                wavelength,
                n1,
                m1,
                domain=mesh,
                degree=degree_source,
            )
            integrand = dot(Escat, X.expression.conj)
            integral = (
                assemble(
                    0.5 * (integrand("+")) * self.dS(boundary) + Constant(0) * self.dx
                )
                / Rcalc**2
            )
            coeff = integral / (sh.sph_hn2(n1, k * Rcalc))
            return coeff

        if coeff_type == "E":
            return _postpro_coeff_e()
        elif coeff_type == "H":
            return _postpro_coeff_h()
        elif coeff_type == "both":
            return _postpro_coeff_e(), _postpro_coeff_h()

    def get_T_matrix(
        self,
        p_max,
        degree_source=None,
        component="transverse",
        boundary="calc_bnds",
    ):
        T11 = np.zeros((p_max, p_max)).tolist()
        T21 = np.zeros((p_max, p_max)).tolist()
        T12 = np.zeros((p_max, p_max)).tolist()
        T22 = np.zeros((p_max, p_max)).tolist()

        solve_again = False
        for p0 in range(1, p_max + 1):
            for p1 in range(1, p_max + 1):
                for source_type in ["M", "N"]:
                    fe, fh = self.get_T_matrix_coeff(
                        p0,
                        p1,
                        source_type,
                        coeff_type="both",
                        solve_again=solve_again,
                        degree_source=degree_source,
                        component=component,
                        boundary=boundary,
                    )
                    solve_again = True

                    if source_type == "M":
                        T11[p0 - 1][p1 - 1] = fh
                        T21[p0 - 1][p1 - 1] = fe
                    else:
                        T12[p0 - 1][p1 - 1] = fh
                        T22[p0 - 1][p1 - 1] = fe

        return [[T11, T12], [T21, T22]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .pw import *


class GaussianBeam(Source):
    """
    GaussianBeam class.

    A Gaussian beam is a common type of beam in optics. It is a solution of the
    paraxial Helmholtz equation, which is a simplified form of the wave equation
    that is only valid for waves propagating in a single direction. The Gaussian
    beam is a good model for the output of a laser, or for a beam that has been
    focused by a lens.

    Parameters
    ----------
    wavelength : float
        The wavelength of the Gaussian beam.
    angle : float
        The angle of incidence of the Gaussian beam.
    waist : float
        The waist of the Gaussian beam, which is the minimum size of the beam.
    position : tuple, optional
        The position of the Gaussian beam. Default is (0, 0).
    dim : int, optional
        The dimension of the Gaussian beam. Default is 2.
    Npw : int, optional
        The number of plane waves used to approximate the Gaussian beam. Default
        is 101.
    phase : float or list of float, optional
        The phase shift of the Gaussian beam. Default is 0.
        If dim is 3, phase can be a list of three floats.
    amplitude : float, optional
        The amplitude of the Gaussian beam. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The Gaussian beam as a dolfin Expression.
    """

    def __init__(
        self,
        wavelength,
        angle,
        waist,
        position=(0, 0),
        dim=2,
        Npw=101,
        phase=0,
        amplitude=1,
        degree=1,
        domain=None,
    ):
        if dim == 3:
            raise NotImplementedError("GaussianBeam not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.angle = angle
        self.waist = waist
        self.position = position
        self.Npw = Npw

    @property
    def expression(self):
        """
        The Gaussian beam as a dolfin Expression.

        The Gaussian beam is modelled as a sum of plane waves with a Gaussian
        distribution of angles.

        Returns
        -------
        expr : Expression
            The Gaussian beam as a dolfin Expression.
        """
        _expression = Constant(0)
        for t in np.linspace(-np.pi / 2, np.pi / 2, self.Npw):
            angle_i = self.angle + t
            K = self.wavenumber * np.array((-np.sin(angle_i), -np.cos(angle_i)))
            position = np.array(self.position)
            phase_pos = -np.dot(K, position)
            term = plane_wave_2d(
                self.wavelength,
                angle_i,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            ) * Constant(
                np.exp(
                    -(t**2) * 4 * (np.pi / 2) ** 2 * (self.waist / self.wavelength) ** 2
                )
            )
            term *= phase_shift_constant(ConstantRe(phase_pos))
            _expression += term
        dk = np.pi / (self.Npw - 1)
        _expression *= Constant(dk)
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import copy
import os

import numpy as np

from . import dolfin
from .complex import *
from .plot import plot


class PML:
    def __init__(
        self,
        direction="x",
        stretch=1 - 1j,
        matched_domain=None,
        applied_domain=None,
    ):
        self.direction = direction
        self.stretch = stretch
        self.matched_domain = matched_domain
        self.applied_domain = applied_domain

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        valid = ["x", "y", "z", "xy", "xz", "yz", "xyz"]
        if value not in valid:
            raise ValueError(f"Unkown direction {value}, must be one of {valid}")

        self._direction = value

    def jacobian_matrix(self):
        if self.direction == "x":
            s = self.stretch, 1, 1
        elif self.direction == "y":
            s = 1, self.stretch, 1
        elif self.direction == "z":
            s = 1, 1, self.stretch
        elif self.direction == "xy":
            s = self.stretch, self.stretch, 1
        elif self.direction == "xz":
            s = self.stretch, 1, self.stretch
        elif self.direction == "yz":
            s = 1, self.stretch, self.stretch
        else:
            s = self.stretch, self.stretch, self.stretch
        return np.diag(s)

    def transformation_matrix(self):
        J = self.jacobian_matrix()
        invJ = np.linalg.inv(J)
        return invJ @ invJ.T * np.linalg.det(J)


class _SubdomainPy(dolfin.UserExpression):
    def __init__(self, markers, subdomains, mapping, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.markers = markers
        self.subdomains = subdomains
        self.mapping = mapping

    def eval_cell(self, values, x, cell):
        for sub, val in self.mapping.items():
            if self.markers[cell.index] == self.subdomains[sub]:
                values[:] = val(x) if callable(val) else val

    def value_shape(self):
        return ()


class _SubdomainCpp(dolfin.CompiledExpression):
    def __init__(self, markers, subdomains, mapping, **kwargs):
        here = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(here, "subdomain.cpp")) as f:
            subdomain_code = f.read()
        compiled_cpp = dolfin.compile_cpp_code(subdomain_code).SubdomainCpp()
        super().__init__(
            compiled_cpp,
            markers=markers,
            subdomains=subdomains,
            mapping=mapping,
            **kwargs,
        )
        self.markers = markers
        self.subdomains = subdomains
        self.mapping = mapping


def _flatten_list(k):
    result = []
    for i in k:
        if isinstance(i, list):
            result.extend(_flatten_list(i))  # Recursive call
        else:
            result.append(i)
    return result


def _separate_mapping_parts(mapping):
    map_re = {}
    map_im = {}
    for k, v in mapping.items():
        try:
            vre = v.real
            vim = v.imag
        except Exception:
            vre = v
            vim = 0 * v
        map_re[k] = vre
        map_im[k] = vim
    return map_re, map_im


def _apply(item, fun):
    return [_apply(x, fun) for x in item] if isinstance(item, list) else fun(item)


def isiter(v):
    return hasattr(v, "__contains__")


def _make_tensor(mapping):
    mapping_tens = mapping.copy()
    N = max(len(v) for v in mapping.values() if isiter(v))
    idf = np.eye(N).tolist()
    for k, v in mapping.items():
        if not isiter(v):

            def fun(idf):
                if idf == 1:
                    idf = v
                return idf

            mapping_tens[k] = _apply(idf, fun)
        else:
            if type(v) == np.ndarray:
                mapping_tens[k] = v.tolist()
    return mapping_tens


def _fldict(k, vflat):
    return [{k: val} for val in vflat]


def _dic2list(dic):
    k = dic.keys()
    val = dic.values()
    N = np.array(list(val)[0]).shape
    nb_keys = len(k)
    L = []
    for k, v in dic.items():
        vflat = _flatten_list(v)
        L.append(_fldict(k, vflat))
    L = np.reshape(L, (nb_keys, np.prod(N))).T
    o = []
    for elem in L:
        d = dict(elem[0])
        for a in elem:
            # d |= a
            d.update(a)
        o.append(d)
    o = np.reshape(o, N).tolist()
    return o


class SubdomainScalarReal:
    def __new__(cls, markers, subdomains, mapping, cpp=True, **kwargs):
        ClassReturn = _SubdomainCpp if cpp else _SubdomainPy
        return ClassReturn(markers, subdomains, mapping, **kwargs)


class SubdomainScalarComplex:
    def __new__(cls, markers, subdomains, mapping, cpp=True, **kwargs):
        mapping_re, mapping_im = _separate_mapping_parts(mapping)
        re = SubdomainScalarReal(markers, subdomains, mapping_re, cpp=cpp, **kwargs)
        im = SubdomainScalarReal(markers, subdomains, mapping_im, cpp=cpp, **kwargs)
        return Complex(re, im)


class SubdomainTensorReal:
    def __new__(self, markers, subdomains, mapping, cpp=True, **kwargs):
        mapping_tensor = _make_tensor(mapping)
        mapping_list = _dic2list(mapping_tensor)

        def fun(mapping_list):
            return SubdomainScalarReal(
                markers, subdomains, mapping_list, cpp=cpp, **kwargs
            )

        q = _apply(mapping_list, fun)
        return dolfin.as_tensor(q)


class SubdomainTensorComplex:
    def __new__(self, markers, subdomains, mapping, cpp=True, **kwargs):
        mapping_tensor = _make_tensor(mapping)
        d = _dic2list(mapping_tensor)
        a = _apply(d, _separate_mapping_parts)
        mape_re = _apply(a, lambda a: a[0])
        mape_im = _apply(a, lambda a: a[1])

        def fun(mapping_list):
            return SubdomainScalarReal(
                markers, subdomains, mapping_list, cpp=cpp, **kwargs
            )

        qre = _apply(mape_re, fun)
        qim = _apply(mape_im, fun)
        Tre = dolfin.as_tensor(qre)
        Tim = dolfin.as_tensor(qim)
        return Complex(Tre, Tim)


class Subdomain:
    def __new__(cls, markers, subdomains, mapping, cpp=True, **kwargs):
        iterable = any(isiter(v) for v in mapping.values())
        flatvals = _flatten_list(mapping.values())
        cplx = any(iscomplex(v) and np.any(v.imag != 0) for v in flatvals)
        if iterable:
            ClassReturn = SubdomainTensorComplex if cplx else SubdomainTensorReal
        else:
            ClassReturn = SubdomainScalarComplex if cplx else SubdomainScalarReal
        return ClassReturn(markers, subdomains, mapping, cpp=cpp, **kwargs)


def tensor_const(T, dim=3, real=False, const=True):
    if dim not in (2, 3):
        raise NotImplementedError("only supports dim = 2 or 3")

    def _treal(T):
        m = []
        for i in range(dim):
            col = []
            for j in range(dim):
                q = dolfin.Constant(T[i][j]) if const else T[i][j]
                col.append(q)
            m.append(col)
        return dolfin.as_tensor(m)

    # assert T.shape == (dim, dim)
    real1 = np.any(
        [[hasattr(t, "real") and hasattr(t, "imag") for t in lines] for lines in T]
    )
    if real1 and not real:
        Treal = [[t.real for t in lines] for lines in T]
        Timag = [[t.imag for t in lines] for lines in T]
        return Complex(_treal(Treal), _treal(Timag))
    else:
        return _treal(T)


def _check_len(p):
    if not hasattr(p, "__len__"):
        return 0
    try:
        lenp = len(p)
    except NotImplementedError:
        lenp = 0
    return lenp


def _make_constant_property_3d(prop, inv=False, real=False):
    new_prop = {}
    for d, p in prop.items():
        lenp = _check_len(p)
        if lenp > 0:
            k = np.linalg.inv(np.array(p)) if inv else np.array(p)
            new_prop[d] = tensor_const(k, dim=3, real=real)
        else:
            k = 1 / p + 0j if inv else p + 0j
            new_prop[d] = k if callable(k) else Constant(k)
    return new_prop


def _make_constant_property_2d(prop, inv=False, real=False):
    new_prop = {}
    for d, p in prop.items():
        lenp = _check_len(p)
        if lenp > 0:
            const = True
            # p = np.array(p)
            p = p[:2][:2]
            for i in range(2):
                for j in range(2):
                    k = p[i][j]
                    if callable(k):
                        const = False
                        break
            # if p.shape != (2, 2):
            #     p = p[:2][:2]
            # k = np.array(p)
            new_prop[d] = tensor_const(p, dim=2, real=real, const=const)
        else:
            k = p + 0j
            new_prop[d] = k if callable(k) else Constant(k)
    return new_prop


def make_constant_property(*args, dim=3, **kwargs):
    if dim == 3:
        return _make_constant_property_3d(*args, **kwargs)
    elif dim == 2:
        return _make_constant_property_2d(*args, **kwargs)
    else:
        raise NotImplementedError("only supports dim = 2 or 3")


def complex_vector(V):
    return Complex(
        dolfin.as_tensor([q.real for q in V]), dolfin.as_tensor([q.imag for q in V])
    )


def _get_xi(prop):
    new_prop = {}
    for d, p in prop.items():
        lenp = _check_len(p)
        if lenp > 0:
            # p = np.array(p)
            k = [[p[0][0], p[1][0]], [p[0][1], p[1][1]]]
            detk = p[0][0] * p[1][1] - p[1][0] * p[0][1]
            k = [[k[0][0] / detk, k[0][1] / detk], [k[1][0] / detk, k[1][1] / detk]]

        else:
            k = 1 / p + 0j
        new_prop[d] = k
    return new_prop


def _get_chi(prop):
    new_prop = {}
    for d, p in prop.items():
        lenp = _check_len(p)
        k = p[2][2] if lenp > 0 else p + 0j
        new_prop[d] = k
    return new_prop


def _invert_3by3_complex_matrix(m):
    m1, m2, m3, m4, m5, m6, m7, m8, m9 = [m[i][j] for i in range(3) for j in range(3)]

    determinant = (
        m1 * m5 * m9
        + m4 * m8 * m3
        + m7 * m2 * m6
        - m1 * m6 * m8
        - m3 * m5 * m7
        - m2 * m4 * m9
    )
    inv = [
        [m5 * m9 - m6 * m8, m3 * m8 - m2 * m9, m2 * m6 - m3 * m5],
        [m6 * m7 - m4 * m9, m1 * m9 - m3 * m7, m3 * m4 - m1 * m6],
        [m4 * m8 - m5 * m7, m2 * m7 - m1 * m8, m1 * m5 - m2 * m4],
    ]
    invre = np.zeros((3, 3), dtype=object)
    invim = np.zeros((3, 3), dtype=object)
    for i in range(3):
        for j in range(3):
            q = inv[i][j]
            invre[i, j] = q.real
            invim[i, j] = q.imag
    invre = invre.tolist()
    invim = invim.tolist()

    return Complex(dolfin.as_tensor(invre), dolfin.as_tensor(invim)) / determinant


def _make_cst_mat(a, b):
    xi = _get_xi(a)
    chi = _get_chi(b)
    xi_ = make_constant_property(xi, dim=2)
    chi_ = make_constant_property(chi, dim=2)
    return xi_, chi_


def _coefs(a, b):
    # xsi = det Q^T/det Q
    def extract(q):
        return dolfin.as_tensor([[q[0][0], q[1][0]], [q[0][1], q[1][1]]])

    def det(M):
        return M[0][0] * M[1][1] - M[1][0] * M[0][1]

    a2 = Complex(extract(a.real), extract(a.imag))
    xi = a2 / det(a2)
    chi = b[2][2]
    return xi, chi


class Coefficient:
    def __init__(self, dict, geometry=None, pmls=None, dim=2, degree=1, element=None):
        if pmls is None:
            pmls = []
        self.dict = dict
        self.geometry = geometry
        self.pmls = pmls
        self.dim = dim
        self.degree = degree
        self.element = element
        if geometry:
            if self.dim == 2:
                markers_key, mapping_key = "triangle", "surfaces"
            else:
                markers_key, mapping_key = "tetra", "volumes"

            self.markers = geometry.mesh_object["markers"][markers_key]
            self.mapping = geometry.subdomains[mapping_key]

        if pmls is not []:
            self.apply_pmls()

    def __repr__(self):
        return f"Coefficient {self.dict.__repr__()}"

    def build_pmls(self):
        new_material_dict = self.dict.copy()
        for pml in self.pmls:
            t = np.array(pml.transformation_matrix())
            eps_pml = (self.dict[pml.matched_domain] * t).tolist()
            new_material_dict[pml.applied_domain] = eps_pml
        return new_material_dict

    def build_annex(self, domains, reference):
        assert reference in self.dict.keys()
        annex_material_dict = self.dict.copy()
        if isinstance(domains, str):
            domains = [domains]
        for dom in domains:
            assert dom in self.dict.keys()
            annex_material_dict[dom] = self.dict[reference]
        annex_material_dict
        annex = copy.copy(self)
        annex.dict = annex_material_dict
        return annex

    def apply_pmls(self):
        self.dict.update(self.build_pmls())

    def as_subdomain(self, **kwargs):
        return Subdomain(
            self.markers,
            self.mapping,
            self.dict,
            degree=self.degree,
            element=self.element,
            **kwargs,
        )

    def as_property(self, dim=None, **kwargs):
        dim = dim or self.dim
        return make_constant_property(self.dict, dim=dim, **kwargs)

    def to_xi(self):
        new = copy.copy(self)
        new.dict = _get_xi(self.dict)
        return new

    def to_chi(self):
        new = copy.copy(self)
        new.dict = _get_chi(self.dict)
        return new

    def plot(self, component=None, **kwargs):
        proj_space = dolfin.FunctionSpace(self.geometry.mesh_object["mesh"], "DG", 0)
        eps_subdomain = self.as_subdomain()
        eps_plot = (
            eps_subdomain
            if component is None
            else eps_subdomain[component[0]][component[1]]
        )
        return plot(eps_plot, proj_space=proj_space, **kwargs)

    def invert(self):
        new = copy.copy(self)
        for dom, val in new.dict.items():
            val_shape = np.array(val).shape
            if val_shape in [(2, 2), (3, 3)]:
                new.dict[dom] = np.linalg.inv(val)
            else:
                new.dict[dom] = 1 / val
        return new



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from .simulation import *


class PhotonicCrystal2D(Simulation):
    """
    2D photonic crystal class

    Parameters
    ----------
    geometry : Lattice2D
        The geometry of the photonic crystal
    epsilon : dict
        Permittivity in various subdomains.
    mu : dict
        Permeability in various subdomains.
    propagation_vector : tuple of float
        The propagation vector of the mode
    boundary_conditions : dict or list of dict
        Boundary conditions of the simulation
    polarization : str
        Polarization of the mode
    degree : int
        The degree of the function space
    eps : float
        The tolerance for the periodic boundary conditions values
    map_tol : float
        The tolerance for the periodic boundary conditions mesh
    """

    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        propagation_vector=(0, 0),
        boundary_conditions=None,
        polarization="TM",
        degree=1,
        eps=dolfin.DOLFIN_EPS,
        map_tol=1e-10,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Lattice2D)
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)
        self.periodic_bcs = BiPeriodic2D(geometry, map_tol=map_tol, eps=eps)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "CG", degree, constrained_domain=self.periodic_bcs
        )
        epsilon_coeff = Coefficient(self.epsilon, geometry, degree=degree)
        mu_coeff = Coefficient(self.mu, geometry, degree=degree)

        coefficients = epsilon_coeff, mu_coeff
        formulation = Maxwell2DBands(
            geometry,
            coefficients,
            function_space,
            propagation_vector=propagation_vector,
            degree=degree,
            polarization=polarization,
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

        self.degree = degree
        self.propagation_vector = propagation_vector

    def eigensolve(self, *args, **kwargs):
        sol = super().eigensolve(*args, **kwargs)
        self.solution["eigenvectors"] = [
            u * self.formulation.phasor for u in sol["eigenvectors"]
        ]
        return self.solution
